import importlib
from typing import Union, Literal
import river
from river.base import DriftDetector, BinaryDriftDetector
from river.drift.binary import EDDM
from river.drift.page_hinkley import PageHinkley
from sail.utils.logging import configure_logger
from sail.common.progress_bar import SAILProgressBar

LOGGER = configure_logger(logger_name="DriftDetector")


class SAILDriftDetector:
    def __init__(
        self,
        model: Union[str, DriftDetector] = EDDM(),
        drift_param: Literal["score", "difference"] = "score",
        verbose=1,
    ) -> None:
        self._drift_detector = self._resolve_drift_detector(model)
        self.drift_param = drift_param
        self.verbose = verbose

    def set_verbose(self, verbose):
        self.verbose = verbose

    def _resolve_drift_detector(self, drift_detector) -> DriftDetector:
        if isinstance(drift_detector, DriftDetector) or isinstance(
            drift_detector, BinaryDriftDetector
        ):
            return drift_detector
        elif isinstance(drift_detector, str):
            if drift_detector == "auto":
                _drift_detector_class = PageHinkley
            elif isinstance(drift_detector, str):
                module = importlib.import_module("river.drift")
                try:
                    _drift_detector_class = getattr(module, drift_detector)
                except AttributeError:
                    raise Exception(
                        f"Drift Detector '{drift_detector}' is not available in River. Available drift detectors: {river.drift.__all__}"
                    )
        else:
            raise TypeError(
                " SAIL Drift Detector `model` must be an instance or str from "
                f"{river.drift.__all__} from river.drift module. Got {drift_detector.__module__}.{drift_detector.__class__.__qualname__}. Set `auto` to use the default."
            )
        return _drift_detector_class()

    def detect_drift(self, score=None, y_pred=None, y_true=None):
        if self.drift_param == "difference":
            return self.detect_drift_with_difference(y_pred, y_true)
        elif self.drift_param == "score":
            return self.detect_drift_with_score(score)

    def detect_drift_with_difference(self, y_preds, y_true):
        with SAILProgressBar(
            steps=len(y_preds),
            desc=f"SAIL Drift detection",
            params={
                "Detector": self._drift_detector.__class__.__qualname__,
                "Batch Size": len(y_preds),
                "Param": "differene",
                "Drift": "No",
            },
            format="scoring",
            verbose=self.verbose,
        ) as progress:
            for yt, yh in zip(y_true, y_preds):
                self._drift_detector.update(yt - yh)
                progress.update()
                if self._drift_detector.drift_detected:
                    progress.update_params("Drift", "Yes")
                    # progress.finish()
                    return True

        return False

    def detect_drift_with_score(self, score):
        with SAILProgressBar(
            steps=1,
            desc=f"SAIL Drift detection",
            params={
                "Detector": self._drift_detector.__class__.__qualname__,
                "Param": "score",
                "Drift": "No",
            },
            format="scoring",
            verbose=self.verbose,
        ) as progress:
            self._drift_detector.update(score)
            progress.update()
            if self._drift_detector.drift_detected:
                progress.update_params("Drift", "Yes")
                return True

        return False



from typing import Literal


class VerboseManager:
    def __init__(
        self,
        name: str = "VerboseManager",
        verbosity: Literal[0, 1] | None = None,
        verbosity_interval: int | None = None,
    ) -> None:
        self.name = name

        assert (
            verbosity == 0 or verbosity == 1 or verbosity is None
        ), "Invalid Verbosity value. Verbosity can only take values from [None, 0, 1]."
        self.verbosity = 0 if verbosity is None else verbosity

        assert (
            isinstance(verbosity_interval, int) or verbosity_interval is None
        ), "Invalid verbosity_interval value. Verbosity interval can only be a None or an int value."
        self.verbosity_interval = verbosity_interval

        self._current_epoch_n = 0
        self._samples_seen_n = 0

    def get_state(self):
        return {
            "current_epoch_n": self._current_epoch_n,
            "samples_seen_n": self._samples_seen_n,
        }

    def set_state(self, state: dict):
        self._current_epoch_n = state["current_epoch_n"]
        self._samples_seen_n = state["samples_seen_n"]

    @property
    def current_epoch_n(self):
        return self._current_epoch_n

    @property
    def samples_seen_n(self):
        return self._samples_seen_n

    def get(self, override=None) -> int:
        if self.if_logging_allowed():
            if override is not None:
                if isinstance(override, int):
                    if override == 0 or override == 1:
                        return override
                else:
                    raise Exception(
                        "Cannot override verbosity value. Override verbosity should only have values from [0, 1]."
                    )
            return self.verbosity
        return 0

    def reset(self) -> int:
        self.verbosity = self.verbosity

    def enabled(self):
        self.verbosity = 1

    def disabled(self):
        self.verbosity = 0

    def if_logging_allowed(self):
        return (
            True
            if self.verbosity_interval is None
            else (
                self._current_epoch_n == 1
                or (self._current_epoch_n % self.verbosity_interval == 0)
            )
        )

    def log_epoch(self, X):
        self._current_epoch_n += 1

        if self.get() == 1:
            self.print_epoch_head()

        self._samples_seen_n += X.shape[0]

    def print_epoch_head(self):
        print(
            f"\n>>> Epoch: {self._current_epoch_n} | Samples Seen: {self._samples_seen_n} -------------------------------------------------------------------------------------"
        )



import numpy as np
from numba import njit, prange
import torch
class RocketModel(torch.nn.Module):
    def __init__(self, input_length, num_kernels):
        super(RocketModel, self).__init__()
        self.input_length = input_length
        self.num_kernels = num_kernels
        self.kernels = self.generate_kernels(self.input_length, self.num_kernels)

    def forward(self, x):
        x_transform = self.apply_kernels(x, self.kernels)
        return x_transform

    #@njit("Tuple((float64[:],int32[:],float64[:],int32[:],int32[:]))(int64,int64)")
    def generate_kernels(self, input_length, num_kernels):

        candidate_lengths = np.array((7, 9, 11), dtype = np.int32)
        lengths = np.random.choice(candidate_lengths, num_kernels)

        weights = np.zeros(lengths.sum(), dtype = np.float64)
        biases = np.zeros(num_kernels, dtype = np.float64)
        dilations = np.zeros(num_kernels, dtype = np.int32)
        paddings = np.zeros(num_kernels, dtype = np.int32)

        a1 = 0

        for i in range(num_kernels):

            _length = lengths[i]

            _weights = np.random.normal(0, 1, _length)

            b1 = a1 + _length
            weights[a1:b1] = _weights - _weights.mean()

            biases[i] = np.random.uniform(-1, 1)

            dilation = 2 ** np.random.uniform(0, np.log2((input_length - 1) / (_length - 1)))
            dilation = np.int32(dilation)
            dilations[i] = dilation

            padding = ((_length - 1) * dilation) // 2 if np.random.randint(2) == 1 else 0
            paddings[i] = padding

            a1 = b1

        return weights, lengths, biases, dilations, paddings


    #@njit("float64[:,:](float64[:,:],Tuple((float64[::1],int32[:],float64[:],int32[:],int32[:])))", parallel = True, fastmath = True)
    def apply_kernels(self, X, kernels):

        weights, lengths, biases, dilations, paddings = kernels

        num_examples, _ = X.shape
        num_kernels = len(lengths)

        _X = np.zeros((num_examples, num_kernels * 2), dtype = np.float64) # 2 features per kernel

        for i in prange(num_examples):

            a1 = 0 # for weights
            a2 = 0 # for features

            for j in range(num_kernels):

                b1 = a1 + lengths[j]
                b2 = a2 + 2

                _X[i, a2:b2] = \
                apply_kernel(X[i], weights[a1:b1], lengths[j], biases[j], dilations[j], paddings[j])

                a1 = b1
                a2 = b2

        return _X

@njit(fastmath = True)
def apply_kernel(X, weights, length, bias, dilation, padding):

    input_length = len(X)

    output_length = (input_length + (2 * padding)) - ((length - 1) * dilation)

    _ppv = 0
    _max = np.NINF

    end = (input_length + padding) - ((length - 1) * dilation)

    for i in range(-padding, end):

        _sum = bias

        index = i

        for j in range(length):

            if index > -1 and index < input_length:

                _sum = _sum + weights[j] * X[index]

            index = index + dilation

        if _sum > _max:
            _max = _sum

        if _sum > 0:
            _ppv += 1

    return _ppv / output_length, _max



# Most of the code from https://github.com/scikit-multiflow/scikit-multiflow/blob/a7e316d/src/skmultiflow/meta/learn_pp.py#L24
import copy
import numpy as np
from sklearn.tree import DecisionTreeClassifier
from sail.models.native import BaseSailObject, ClassifierMixin, MetaEstimatorMixin
import warnings
from sklearn.utils import check_random_state


def LearnPP(
    base_estimator=DecisionTreeClassifier(),
    error_threshold=0.5,
    n_estimators=30,
    n_ensembles=10,
    window_size=100,
    random_state=None,
):  # pragma: no cover
    warnings.warn(
        "'LearnPP' has been renamed to 'LearnPPClassifier' in v0.5.0.\n"
        "The old name will be removed in v0.7.0",
        category=FutureWarning,
    )
    return LearnPPClassifier(
        base_estimator=base_estimator,
        error_threshold=error_threshold,
        n_estimators=n_estimators,
        n_ensembles=n_ensembles,
        window_size=window_size,
        random_state=random_state,
    )


class LearnPPClassifier(BaseSailObject, ClassifierMixin, MetaEstimatorMixin):
    """Learn++ ensemble classifier.

    Learn++ [1]_  does not require access to previously used data during
    subsequent incremental learning steps. At the same time, it does not
    forget previously acquired knowledge. Learn++ utilizes an ensemble of
    classifiers by generating multiple hypotheses using training data
    sampled according to carefully tailored distributions.

    References
    ----------
    .. [1] Polikar, Robi and Upda, Lalita and Upda, Satish S and Honavar, Vasant.
       Learn++: An Incremental Learning Algorithm for Supervised Neural Networks.
       IEEE Transactions on Systems Man and Cybernetics Part C (Applications and Reviews), 2002.

    Parameters
    ----------
    base_estimator: sklearn.BaseEstimator (default=DecisionTreeClassifier)
        Each member of the ensemble is an instance of the base estimator.
    n_estimators: int (default=30)
        The number of classifiers per ensemble
    n_ensembles: int (default=10)
        The number of ensembles to keep.
    window_size: int (default=100)
        The size of the training window (batch), in other words, how many instances are kept for training.
    error_threshold: float (default=0.5)
        Only keep the learner with the error smaller than error_threshold
    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by `np.random`.

    Raises
    ------
    RuntimeError:
        A RuntimeError is raised if the base_estimator is too weak. In other words,
        it has too low accuracy on the dataset. A RuntimeError is also raised if the
        'classes' parameter is not passed in the first partial_fit call, or if they
        are passed in further calls but differ from the initial classes.

    Notes
    -----
        Originally, Learn++ is designed to train all of its members and combine their predictions
        considering the observed normalized errors. However, when training the base estimators, if
        the observed prediction error shrinks to zero before all estimators are trained, the error
        normalization is ill-defined, i.e., the instance error-based weight normalization factor
        (the sum of the errors) is zero. This implementation adds an 'early stop' mechanism to
        circumvent this corner case: LearnPPClassifier stops adding members to the ensemble if all
        instances are correctly classified. Otherwise (its normal behavior), the model uses as many
        ensemble members as defined via the ``n_estimators`` parameter.

    Examples
    --------
    Learn++ classifier performance: 0.9555

    """

    def __init__(
        self,
        base_estimator=DecisionTreeClassifier(),
        error_threshold=0.5,
        n_estimators=30,
        n_ensembles=10,
        window_size=100,
        random_state=None,
    ):
        super().__init__()
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.ensembles = []
        self.ensemble_weights = []
        self.classes = None
        self.n_ensembles = n_ensembles
        self.random = check_random_state(random_state)
        self.random_state = random_state
        self.error_threshold = error_threshold
        self.X_batch = []
        self.y_batch = []
        self.window_size = window_size

    def reset(self):
        self.ensembles = []
        self.ensemble_weights = []
        self.X_batch = []
        self.y_batch = []
        self.random = check_random_state(self.random_state)

    def partial_fit(self, X, y, classes=None, sample_weight=None):
        """Partially (incrementally) fit the model.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.

        y: numpy.ndarray of shape (n_samples)
            An array-like with the class labels of all samples in X.

        classes: numpy.ndarray, optional (default=None)
            Array with all possible/known class labels. This is an optional parameter, except
            for the first partial_fit call where it is compulsory.

        sample_weight: NOT used (default=None)

        Raises
        ------
        RuntimeError:
            A RuntimeError is raised if the 'classes' parameter is not
            passed in the first partial_fit call, or if they are passed in further
            calls but differ from the initial classes list passed.
            A RuntimeError is raised if the base_estimator is too weak. In other word,
            it has too low accuracy on the dataset.

        Returns
        -------
        LearnPPClassifier
            self
        """

        if self.classes is None:
            if classes is None:
                raise RuntimeError(
                    "Should pass the classes in the first partial_fit call"
                )
            else:
                self.classes = classes

        if classes is not None and self.classes is not None:
            if set(classes) == set(self.classes):
                pass
            else:
                raise RuntimeError("The values of classes are different")

        N, _ = X.shape

        for i in range(N):
            self.X_batch.append(X[i])
            self.y_batch.append(y[i])
            if len(self.y_batch) == self.window_size:
                self.__fit_batch(self.X_batch, self.y_batch)
                self.X_batch = []
                self.y_batch = []

        return self

    def __fit_batch(self, X, y):
        ensemble = []
        normalized_errors = []

        m = len(X)
        X = np.array(X)
        y = np.array(y)

        Dt = np.ones((m,)) / m

        items_index = np.linspace(0, m - 1, m)
        t = 0
        while t < self.n_estimators:
            patience = 0

            # Set distribution Dt
            sum_dt = np.sum(Dt)
            if sum_dt == 0:  # Early stop in case all instances are correctly classified
                break
            Dt = Dt / sum_dt

            total_error = 1.0
            while total_error >= self.error_threshold:
                # create training and testing subsets according to Dt
                train_size = int(m / 2)
                test_size = int(m / 2)
                train_items_index = self._get_item(items_index, Dt, train_size)
                test_items_index = self._get_item(items_index, Dt, test_size)

                X_train = X[train_items_index]
                y_train = y[train_items_index]
                X_test = X[test_items_index]
                y_test = y[test_items_index]

                # Train a weak learner
                if t > len(ensemble) - 1:
                    ensemble.append(copy.deepcopy(self.base_estimator))
                    normalized_errors.append(1.0)
                try:
                    ensemble[t].fit(X_train, y_train)
                except NotImplementedError:
                    ensemble[t].partial_fit(X_train, y_train)

                # predict on the data
                y_predict = ensemble[t].predict(X_test)

                total_error = self.__compute_error(
                    Dt[test_items_index], y_test, y_predict
                )

                if total_error < self.error_threshold:
                    norm_error = total_error / (1 - total_error)
                    normalized_errors[t] = norm_error

                    # predict using all hypothesis in the ensemble with majority votes
                    y_predict_composite = self.__majority_vote(
                        X, t + 1, ensemble, normalized_errors
                    )

                    total_error = self.__compute_error(Dt, y, y_predict_composite)
                    if total_error < self.error_threshold:
                        normalize_composite_error = total_error / (1 - total_error)
                        if t < self.n_estimators - 1:
                            Dt[y_predict_composite == y] = (
                                Dt[y_predict_composite == y] * normalize_composite_error
                            )

                if total_error > self.error_threshold:
                    patience += 1
                else:
                    patience = 0
                if patience > 1000:
                    raise RuntimeError("Your base estimator is too weak")
            t += 1

        self.ensembles.append(ensemble)
        self.ensemble_weights.append(normalized_errors)

        if len(self.ensembles) > self.n_ensembles:
            self.ensembles.pop(0)
            self.ensemble_weights.pop(0)

        return self

    @staticmethod
    def __compute_error(Dt, y_true, y_predict):
        total_error = np.sum(Dt[y_predict != y_true]) / np.sum(Dt)
        return total_error

    def __vote_proba(self, X, t, ensemble, normalized_errors):
        res = []
        for m in range(len(X)):
            votes = np.zeros(len(self.classes))
            for i in range(t):
                h = ensemble[i]

                y_predicts = h.predict(X[m].reshape(1, -1))
                norm_error = normalized_errors[i]
                votes[int(y_predicts[0])] += np.log(1 / (norm_error + 1e-50))

            res.append(votes)
        return res

    def __majority_vote(self, X, t, ensemble, normalized_errors):
        res = self.__vote_proba(X, t, ensemble, normalized_errors)
        return np.argmax(res, axis=1)

    def predict_proba(self, X):
        """Predicts the probability of each sample belonging to each one of the
        known classes.

        Parameters
        ----------
        X: numpy.ndarray of shape (n_samples, n_features)
            A matrix of the samples we want to predict.

        Returns
        -------
        numpy.ndarray
            An array of shape (n_samples, n_features), in which each outer entry is
            associated with the X entry of the same index. And where the list in
            index [i] contains len(self.target_values) elements, each of which represents
            the probability that the i-th sample of X belongs to a certain label.

        """
        votes = np.zeros((len(X), len(self.classes)))
        for i in range(len(self.ensembles)):
            ensemble = self.ensembles[i]
            ensemble_weight = self.ensemble_weights[i]
            votes += np.array(
                self.__vote_proba(X, len(ensemble), ensemble, ensemble_weight)
            )
        return votes

    def predict(self, X):
        """Predict classes for the passed data.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of data samples to predict the labels for.

        Returns
        -------
        A numpy.ndarray with all the predictions for the samples in X.

        Notes
        -----

        The predict function uses majority votes from all its learners
        with their weights to find the most likely prediction for the sample matrix X.

        """
        votes = self.predict_proba(X)
        return np.argmax(votes, axis=1)

    def _get_item(self, items, items_weights, number_of_items):
        return self.random.choice(items, number_of_items, p=items_weights).astype(
            np.int32
        )



import numpy as np
from numpy.linalg import pinv
import pandas as pd
from sail.models.native.base import BaseEstimator

# Code and algorithms used as reference to convert to sklearn/sail template:
# https://medium.datadriveninvestor.com/extreme-learning-machine-for-simple-classification-e776ad797a3c
# https://github.com/chickenbestlover/Online-Recurrent-Extreme-Learning-Machine/blob/master/timeseries_prediction.ipynb
# https://github.com/khushalpt/Incremental-Extreme-Machine-Learning/blob/main/IELM_Regression_UCI.ipynb
# https://github.com/5663015/elm/blob/master/elm.py


def _sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))


def _relu(x):
    return np.maximum(0, x)


def _linear(X, weights, bias):
    return np.dot(X, np.transpose(weights)) + bias


class IELM(BaseEstimator):
    def __init__(
        self,
        numInputNodes,
        numOutputNodes,
        numHiddenNodes,
        forgettingFactor=0.999,
        lambda_value=0.01,
    ):
        """
        :param numInputNodes: int. Number of input nodes.
        :param numOutputNodes: int. Number of output nodes.
        :param numHiddenNodes: int. Number of hidden nodes.
        :param forgettingFactor: float, 0<=forgettingFactor<=1. Forgetting factor for old samples.
        :param lambda_value: float. Regularization factor.
        """

        self._inputNodes = numInputNodes
        self._outputNodes = numOutputNodes
        self._numHiddenNodes = numHiddenNodes

        self._weights = np.random.random((self._numHiddenNodes, self._inputNodes))
        self._weights = self._weights * 2 - 1

        self._bias = np.random.random((1, self._numHiddenNodes)) * 2 - 1

        self._beta = np.zeros([self._numHiddenNodes, self._outputNodes])

        self._M = np.linalg.inv(lambda_value * np.eye(self._numHiddenNodes))

        self.forgettingFactor = forgettingFactor

    def calculateHiddenLayerActivation(self, X):
        """
        :param X : numpy array of shape (n_samples, n_features)
            The input samples.
        :return: numpy array of shape (n_samples, n_features).
             Transformed sample
        """
        V = _linear(X, self._weights, self._bias)
        # H = sigmoidActFunc(V)
        H = _relu(V)
        return H

    def fit(self, X, y):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
                   Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return: self
        """
        # if isinstance(y, pd.Series):
        #     y = (y.values).reshape(-1)
        numSamples = None
        numOutputs = None
        if y.ndim == 1:
            y = y.to_numpy()  # np.reshape(y.values,-1)
            y = y.reshape((y.shape[0], 1))
            (numSamples, numOutputs) = (y.shape[0], 1)
        else:
            (numSamples, numOutputs) = y.shape
        assert X.shape[0] == y.shape[0]

        H = self.calculateHiddenLayerActivation(X)
        Ht = np.transpose(H)

        self._M = (1 / self.forgettingFactor) * self._M - np.dot(
            (1 / self.forgettingFactor) * self._M,
            np.dot(
                Ht,
                np.dot(
                    pinv(
                        np.eye(numSamples)
                        + np.dot(H, np.dot((1 / self.forgettingFactor) * self._M, Ht))
                    ),
                    np.dot(H, (1 / self.forgettingFactor) * self._M),
                ),
            ),
        )

        self._beta = self._beta + np.dot(self._M, np.dot(Ht, y - np.dot(H, self._beta)))
        return self

    def predict(self, X):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
           Input samples.
        :return: numpy.ndarray of shape (n_samples)
            Predictions for the samples in X.
        """
        H = self.calculateHiddenLayerActivation(X)
        prediction = np.dot(H, self._beta)
        return prediction.ravel()

    def partial_fit(self, X, y, **kwargs):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
                   Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return: self
        """
        return self.fit(X, y)



import os
import shutil
from typing import Any, List, Literal, Tuple

import numpy as np
import pandas as pd
from sklearn import utils
from sklearn.base import clone
from sklearn.pipeline import Pipeline
from sklearn.utils import _print_elapsed_time
from sklearn.utils.metaestimators import available_if

from sail.common.decorators import log_epoch
from sail.common.helper import VerboseManager
from sail.utils.logging import configure_logger
from sail.common.progress_bar import SAILProgressBar
from sail.common.scorer import SAILModelScorer
from sail.utils.serialization import load_obj, save_obj

LOGGER = configure_logger(logger_name="SAILPipeline")


class SAILPipeline(Pipeline):
    def __init__(
        self,
        steps: List[Tuple[str, Any]],
        scoring=None,
        cache_memory=None,
        verbosity_level: Literal[0, 1] | None = 1,
        verbosity_interval: int | None = None,
    ):
        """[summary]

        Args:
            steps (list, optional): List of steps for the pipeline, required. Each step is a tuple containing a string, representing the name of a step, and a python object for the step.
            scoring: pipeline score metric
            verbosity: verbosity level for training logs. 0 (No logs) is default.
        """
        super(SAILPipeline, self).__init__(steps, memory=cache_memory, verbose=False)
        self.scoring = scoring
        self.cache_memory = cache_memory
        self.verbosity_level = verbosity_level
        self.verbosity_interval = verbosity_interval

        # validate and create verbosity manager
        self.verbosity = self._resolve_verbosity(verbosity_level, verbosity_interval)

        # validate and create scorer
        self._scorer = self._validate_and_get_scorer(scoring, steps[-1][1])

    def _can_fit_transform(self):
        return self._final_estimator == "passthrough" or (
            hasattr(self._final_estimator, "fit")
            and hasattr(self._final_estimator, "transform")
        )

    def _can_parital_fit_transform(self):
        return self._final_estimator == "passthrough" or (
            hasattr(self._final_estimator, "partial_fit")
            and hasattr(self._final_estimator, "transform")
        )

    def _resolve_verbosity(self, verbosity, verbosity_interval):
        if verbosity is None:
            return VerboseManager(
                "SAILPipeline",
                0,
                verbosity_interval,
            )
        elif isinstance(verbosity, int):
            return VerboseManager(
                "SAILPipeline",
                verbosity,
                verbosity_interval,
            )

    def _validate_and_get_scorer(self, scoring, estimator):
        estimator_type = None

        if estimator == "passthrough":
            estimator_type = estimator
        elif hasattr(estimator, "_estimator_type"):
            estimator_type = estimator._estimator_type

        if estimator_type:
            estimator_type = SAILModelScorer(
                scoring=scoring,
                estimator_type=estimator_type,
                pipeline_mode=True,
            )

        return estimator_type

    @property
    def get_progressive_score(self):
        if self._scorer:
            return self._scorer.get_progressive_score

    def score(self, X, y, sample_weight=1.0, verbose: Literal[0, 1] | None = None):
        if self._scorer:
            y_preds = self.predict(X)
            if isinstance(y, pd.Series):
                y = y.to_numpy()
            return self._scorer.score(
                y, y_preds, sample_weight, verbose=self.verbosity.get(verbose)
            )

    def score_estimator(
        self, X, y, sample_weight: float = 1.0, verbose: Literal[0, 1] | None = None
    ):
        if self._scorer:
            y_preds = self._final_estimator.predict(X)
            if isinstance(y, pd.Series):
                y = y.to_numpy()
            return self._scorer.score(
                y, y_preds, sample_weight, verbose=self.verbosity.get(verbose)
            )

    def _progressive_score(
        self,
        X,
        y,
        sample_weight=1.0,
        detached=False,
        verbose: Literal[0, 1] | None = None,
    ):
        if self._scorer:
            y_pred = self.predict(X)
            if isinstance(y, pd.Series):
                y = y.to_numpy()
            return self._scorer.progressive_score(
                y, y_pred, sample_weight, detached, self.verbosity.get(verbose)
            )

    @log_epoch
    def fit(self, X, y=None, **fit_params):
        self._fit(X, y, warm_start=False, **fit_params)

    @log_epoch
    def partial_fit(self, X, y=None, **fit_params):
        self._partial_fit(X, y, **fit_params)

    def _partial_fit(self, X, y=None, **fit_params):
        if self.__sklearn_is_fitted__():
            self._progressive_score(X, y, verbose=0)
        self._fit(X, y, warm_start=True, **fit_params)

    def _fit(self, X, y=None, warm_start: bool = None, **fit_params):
        fit_params_steps = self._check_fit_params(**fit_params)
        # shallow copy of steps - this should really be steps_
        self.steps = list(self.steps)
        self._validate_steps()
        # Setup the memory
        memory = utils.validation.check_memory(self.memory)

        def _fit_transform_one(transformer, X, y, **fit_params):
            if warm_start and hasattr(transformer, "partial_fit"):
                transformed_X = transformer.partial_fit(X, y, **fit_params).transform(X)
            elif hasattr(transformer, "fit_transform"):
                transformed_X = transformer.fit_transform(X, y, **fit_params)
            else:
                transformed_X = transformer.fit(X, y, **fit_params).transform(X)
            return transformed_X, transformer

        fit_transform_one_cached = memory.cache(_fit_transform_one)

        with SAILProgressBar(
            steps=len(list(self._iter(with_final=True, filter_passthrough=True))),
            desc=f"SAIL Pipeline Partial fit" if warm_start else f"SAIL Pipeline fit",
            params={
                "Batch Size": X.shape[0],
            },
            format="pipeline_training",
            verbose=self.verbosity.get(),
        ) as progress:
            for step_idx, name, transformer in self._iter(
                with_final=False, filter_passthrough=True
            ):
                progress.append_desc(f"[{name}]")
                if transformer is None or transformer == "passthrough":
                    with _print_elapsed_time("Pipeline", self._log_message(step_idx)):
                        continue

                if hasattr(memory, "location") and memory.location is None:
                    # we do not clone when caching is disabled to
                    # preserve backward compatibility
                    cloned_transformer = transformer
                else:
                    cloned_transformer = clone(transformer)
                # Fit or load from cache the current transformer
                X, fitted_transformer = fit_transform_one_cached(
                    cloned_transformer,
                    X,
                    y,
                    **fit_params_steps[name],
                )
                # Replace the transformer of the step with the fitted
                # transformer. This is necessary when loading the transformer
                # from the cache.
                self.steps[step_idx] = (name, fitted_transformer)

                progress.update()

            # call fit / partial_fit on the final estimator
            with _print_elapsed_time(
                "SAIL-FinalEstimator", self._log_message(len(self.steps) - 1)
            ):
                if self._final_estimator != "passthrough":
                    progress.append_desc(f"[{self.steps[-1][0]}]")
                    self.fit_final_estimator(
                        X, y, warm_start=warm_start, verbose=0, **fit_params_steps
                    )
                progress.update()

            # Update progress bar score after fit().
            if self.verbosity.get():
                if warm_start:
                    progress.update_params("P_Score", self.get_progressive_score)
                else:
                    progress.update_params(
                        "Score", self.score_estimator(X, y, verbose=0)
                    )

        return self

    def fit_final_estimator(
        self,
        X,
        y,
        warm_start: bool = False,
        verbose: Literal[0, 1] | None = None,
        **fit_params_steps,
    ):
        verbose = self.verbosity.get(verbose)
        with SAILProgressBar(
            steps=1,
            desc=f"SAIL Model Partial fit" if warm_start else f"SAIL Model fit",
            params={
                "Model": self.steps[-1][1].__class__.__name__,
                "Batch Size": X.shape[0],
            },
            format="model_training",
            verbose=verbose,
        ) as progress:
            if self.steps[-1][0] in fit_params_steps:
                fit_params_last_step = fit_params_steps[self.steps[-1][0]]
            else:
                fit_params_last_step = {}

            if warm_start:
                if not hasattr(self._final_estimator, "partial_fit"):
                    raise AttributeError(
                        f"Final Estimator: '{self.steps[-1][0]}' does not implement partial_fit()."
                    )
                self._final_estimator.partial_fit(X, y, **fit_params_last_step)

                # Update progress bar
                if verbose == 1:
                    progress.update()
                    progress.update_params("P_Score", self.get_progressive_score)
            else:
                if not hasattr(self._final_estimator, "fit"):
                    raise AttributeError(
                        f"Final Estimator '{self.steps[-1][0]}' does not implement fit()."
                    )
                fit_params_last_step.pop("classes", None)
                self._final_estimator.fit(X, y, **fit_params_last_step)

                # Update progress bar
                if verbose == 1:
                    progress.update()
                    progress.update_params(
                        "Score", self.score_estimator(X, y, verbose=0)
                    )

    @available_if(_can_fit_transform)
    def fit_transform(self, X, y=None, **fit_params):
        """Fit the model and transform with the final estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed samples.
        """
        self.fit(X, y, **fit_params)
        return self.transform(X)

    @available_if(_can_parital_fit_transform)
    def partial_fit_transform(self, X, y=None, **fit_params):
        """Partial Fit the model and transform with the final estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed samples.
        """
        self.partial_fit(X, y, **fit_params)
        return self.transform(X)

    def save(self, model_folder, name="sail_pipeline", overwrite=True) -> str:
        """
        Parameters:
        -----------
        model_folder: str
            location to save the model
        overwrite: bool (False)
            if True and model_folder already exists, it will be delted and recreated

        Returns
        -------
        saved_location: str
        """
        save_location = os.path.join(model_folder, name)
        if not overwrite and os.path.exists(save_location):
            raise Exception(
                f"{save_location} already exists, specify overwrite=True to replace contents"
            )
        else:
            # delete old folder to avoid unwanted file overlaps
            if os.path.exists(save_location) and os.path.isdir(save_location):
                shutil.rmtree(save_location)

            LOGGER.debug(f"making directory tree {save_location}")
            os.makedirs(save_location, exist_ok=True)
            if not os.path.exists(save_location):
                raise Exception(f"target directory {save_location} can not be created!")

        # --------------------------------------------
        # explicity save all the steps and steps names
        # --------------------------------------------
        for i, (step_name, step) in enumerate(self.steps):
            save_obj(
                obj=step,
                location=os.path.join(save_location, "steps"),
                file_name=step_name,
            )
        save_obj(
            obj=[step_name for (step_name, _) in self.steps],
            location=os.path.join(save_location, "steps"),
            file_name="steps_meta",
            serialize_type="json",
        )

        # -----------------------------
        # save scorer progressive state
        # -----------------------------
        state = self._scorer.get_state()
        if state:
            np.savez(
                os.path.join(save_location, "scorer_state"),
                y_true=state["y_true"],
                y_pred=state["y_pred"],
            )

        # -------------------------------------------
        # save verbosity state
        # -------------------------------------------
        save_obj(
            obj=self.verbosity.get_state(),
            location=os.path.join(save_location),
            file_name="verbosity_state",
            serialize_type="json",
        )

        # -------------------------------------------
        # save rest of the params
        # -------------------------------------------
        params = self.get_params(deep=False)
        params.pop("steps")
        save_obj(
            obj=params,
            location=save_location,
            file_name="params",
        )

        return save_location

    @classmethod
    def load(cls, model_folder, name="sail_pipeline"):
        load_location = os.path.join(model_folder, name)

        # -------------------------------------------
        # Load steps to add to sail pipeline
        # -------------------------------------------
        steps = []
        steps_location = os.path.join(load_location, "steps")
        steps_meta = load_obj(
            location=steps_location,
            file_name="steps_meta",
            serialize_type="json",
        )
        for step_name in steps_meta:
            step_obj = load_obj(location=steps_location, file_name=step_name)
            steps.append((step_name, step_obj))

        # -------------------------------------------
        # Load params
        # -------------------------------------------
        params = load_obj(location=load_location, file_name="params")

        # -------------------------------------------
        # create pipeline
        # -------------------------------------------
        sail_pipeline = SAILPipeline(steps=steps, **params)

        # -------------------------------------------
        # Pre-load progressive scorer state from initial_points if exist.
        # -------------------------------------------
        if os.path.exists(os.path.join(load_location, "scorer_state.npz")):
            initial_points = np.load(os.path.join(load_location, "scorer_state.npz"))
            sail_pipeline._scorer.set_state(initial_points)
            # sail_pipeline._scorer.progressive_score(
            #     initial_points["y_true"], initial_points["y_pred"], verbose=1
            # )
            initial_points.close()

        # -------------------------------------------
        # load verbosity state
        # -------------------------------------------
        state = load_obj(
            location=load_location,
            file_name="verbosity_state",
            serialize_type="json",
        )
        sail_pipeline.verbosity.set_state(state)

        return sail_pipeline



from sail.models.auto_ml.base_strategy import (
    PipelineActionType,
    PipelineStrategy,
    PipelineActions,
)
from sail.utils.logging import configure_logger

LOGGER = configure_logger(logger_name="PipelineStrategy")


class DetectAndIncrement(PipelineStrategy):
    def __init__(
        self,
        search_method,
        search_data_size,
        drift_detector,
        **kwargs,
    ) -> None:
        super(DetectAndIncrement, self).__init__(
            search_method, search_data_size, drift_detector, **kwargs
        )

        # Add all pipeline actions
        self.pipeline_actions = PipelineActions()
        self.pipeline_actions.add_action(PipelineActionType.DATA_COLLECTION)
        self.pipeline_actions.add_action(PipelineActionType.FIND_BEST_PIPELINE)
        self.pipeline_actions.add_action(PipelineActionType.SCORE_AND_DETECT_DRIFT)
        self.pipeline_actions.add_action(
            PipelineActionType.PARTIAL_FIT_MODEL,
            next=PipelineActionType.SCORE_AND_DETECT_DRIFT,
        )

        LOGGER.info(
            f"Pipeline Strategy [{self.__class__.__name__}] created with actions: {self.pipeline_actions.get_actions()}"
        )

    def detect_drift(self, **kwargs):
        if self.drift_detector.detect_drift(**kwargs):
            if self.verbosity.get() == 0:
                print(
                    f"\n>>> Epoch: {self.verbosity.current_epoch_n} | Samples Seen: {self.verbosity.samples_seen_n} -------------------------------------------------------------------------------------"
                )
            LOGGER.info(
                "Drift Detected in the data. Final Estimator will be incrementally trained on the next train()"
            )
            self.pipeline_actions.next()
            return True
        return False


class DetectAndRetrain(PipelineStrategy):
    def __init__(
        self,
        search_method,
        search_data_size,
        drift_detector,
        **kwargs,
    ) -> None:
        super(DetectAndRetrain, self).__init__(
            search_method, search_data_size, drift_detector, **kwargs
        )

        # Add all pipeline actions
        self.pipeline_actions = PipelineActions()
        self.pipeline_actions.add_action(PipelineActionType.DATA_COLLECTION)
        self.pipeline_actions.add_action(PipelineActionType.FIND_BEST_PIPELINE)
        self.pipeline_actions.add_action(PipelineActionType.SCORE_AND_DETECT_DRIFT)
        self.pipeline_actions.add_action(
            PipelineActionType.FIT_MODEL,
            next=PipelineActionType.SCORE_AND_DETECT_DRIFT,
        )

        LOGGER.info(
            f"Pipeline Strategy [{self.__class__.__name__}] created with actions: {self.pipeline_actions.get_actions()}"
        )

    def detect_drift(self, **kwargs):
        if self.drift_detector.detect_drift(**kwargs):
            if self.verbosity.get() == 0:
                self.verbosity.print_epoch_head()
            LOGGER.info(
                "Drift Detected in the data. Final Estimator will be re-trained on the next train()"
            )
            self.pipeline_actions.next()
            return True
        return False


class DetectAndWarmStart(PipelineStrategy):
    def __init__(
        self,
        search_method,
        search_data_size,
        drift_detector,
        **kwargs,
    ) -> None:
        super(DetectAndWarmStart, self).__init__(
            search_method, search_data_size, drift_detector, **kwargs
        )

        # Add all pipeline actions
        self.pipeline_actions = PipelineActions()
        self.pipeline_actions.add_action(PipelineActionType.DATA_COLLECTION)
        self.pipeline_actions.add_action(PipelineActionType.FIND_BEST_PIPELINE)
        self.pipeline_actions.add_action(PipelineActionType.SCORE_AND_DETECT_DRIFT)
        self.pipeline_actions.add_action(PipelineActionType.DATA_COLLECTION)
        self.pipeline_actions.add_action(
            PipelineActionType.WARM_START_FIND_BEST_PIPELINE,
            next=PipelineActionType.SCORE_AND_DETECT_DRIFT,
        )
        LOGGER.info(
            f"Pipeline Strategy [{self.__class__.__name__}] created with actions: {self.pipeline_actions.get_actions()}"
        )

    def detect_drift(self, **kwargs):
        if self.drift_detector.detect_drift(**kwargs):
            if self.verbosity.get() == 0:
                self.verbosity.print_epoch_head()
            LOGGER.info(
                "Drift Detected in the data. SAIL AutoML will re-start with previously evaluated configurations on the next train()"
            )
            self.pipeline_actions.next()
            return True
        return False


class DetectAndRestart(PipelineStrategy):
    def __init__(
        self,
        search_method,
        search_data_size,
        drift_detector,
        **kwargs,
    ) -> None:
        super(DetectAndRestart, self).__init__(
            search_method, search_data_size, drift_detector, **kwargs
        )

        # Add all pipeline actions
        self.pipeline_actions = PipelineActions()
        self.pipeline_actions.add_action(PipelineActionType.DATA_COLLECTION)
        self.pipeline_actions.add_action(PipelineActionType.FIND_BEST_PIPELINE)
        self.pipeline_actions.add_action(
            PipelineActionType.SCORE_AND_DETECT_DRIFT,
            next=PipelineActionType.DATA_COLLECTION,
        )

        LOGGER.info(
            f"Pipeline Strategy [{self.__class__.__name__}] created with actions: {self.pipeline_actions.get_actions()}"
        )

    def detect_drift(self, **kwargs):
        if self.drift_detector.detect_drift(**kwargs):
            if self.verbosity.get() == 0:
                self.verbosity.print_epoch_head()
            LOGGER.info(
                "Drift Detected in the data. SAIL AutoML will re-start from scratch on the next train()"
            )
            self.pipeline_actions.next()
            return True
        return False


class PeriodicRestart(PipelineStrategy):
    def __init__(
        self,
        search_method,
        search_data_size,
        drift_detector,
        **kwargs,
    ) -> None:
        super(PeriodicRestart, self).__init__(
            search_method, search_data_size, drift_detector, **kwargs
        )

        # Add all pipeline actions
        self.pipeline_actions = PipelineActions()
        self.pipeline_actions.add_action(PipelineActionType.DATA_COLLECTION)
        self.pipeline_actions.add_action(
            PipelineActionType.FIND_BEST_PIPELINE,
            next=PipelineActionType.DATA_COLLECTION,
        )

        LOGGER.info(
            f"Pipeline Strategy [{self.__class__.__name__}] created with actions: {self.pipeline_actions.get_actions()}"
        )


class PrequentialTraining(PipelineStrategy):
    def __init__(
        self,
        search_method,
        search_data_size,
        **kwargs,
    ) -> None:
        kwargs.pop("incremental_training")
        super(PrequentialTraining, self).__init__(
            search_method,
            search_data_size,
            incremental_training=True,
            **kwargs,
        )

        # Add all pipeline actions
        self.pipeline_actions = PipelineActions()
        self.pipeline_actions.add_action(PipelineActionType.DATA_COLLECTION)
        self.pipeline_actions.add_action(PipelineActionType.FIND_BEST_PIPELINE)
        self.pipeline_actions.add_action(
            PipelineActionType.PARTIAL_FIT_PIPELINE,
            next=PipelineActionType.PARTIAL_FIT_PIPELINE,
        )

        LOGGER.info(
            f"Pipeline Strategy [{self.__class__.__name__}] created with actions: {self.pipeline_actions.get_actions()}"
        )



import torch
import torch.nn as nn
from skorch.regressor import NeuralNetRegressor
from sail.models.torch.base import TorchSerializationMixin


class _RNNModel(nn.Module):
    """Basic RNN/GRU model.

    Args:
        input_units (int): Number of input units
        output_units (int): Number of output units
        hidden_units (int): Number of hidden units
        n_hidden_layers (int): Number of hidden layers
        output_nonlin (torch.nn.Module instance or None (default=nn.Linear)):
            Non-linearity to apply after last layer, if any.
        dropout (float): Dropout
        squeeze_output (bool): default=False
            Whether to squeeze output. For Skorch consistency.
            https://github.com/skorch-dev/skorch/blob/master/skorch/toy.py
        cell_type (string): default="RNN"
    """

    def __init__(
        self,
        input_units,
        output_units,
        hidden_units,
        n_hidden_layers=1,
        dropout=0.2,
        output_nonlin=nn.Linear,
        squeeze_output=False,
        cell_type="RNN",
    ):
        super(_RNNModel, self).__init__()
        self.input_units = input_units
        self.output_units = output_units
        self.hidden_units = hidden_units
        self.n_hidden_layers = n_hidden_layers
        self.output_nonlin = output_nonlin
        self.squeeze_output = squeeze_output
        if cell_type == "RNN":
            rnn = nn.RNNCell
        elif cell_type == "GRU":
            rnn = nn.GRUCell
        else:
            raise ValueError(
                f"RNN type {cell_type} is not supported. supported: [RNN, GRU]"
            )

        self.rnns = nn.ModuleList(
            [rnn(self.input_units, self.hidden_units)]
            + [
                rnn(self.hidden_units, self.hidden_units)
                for i in range(self.n_hidden_layers - 1)
            ]
        )

        if self.output_nonlin:
            self.out = self.output_nonlin(self.hidden_units, self.output_units)
        self.do = nn.Dropout(p=dropout)
        self.actfn = nn.Tanh()
        self.device = torch.device("cpu")
        self.dtype = torch.float

    def forward(self, x, h0=None, train=False):

        hs = x  # initiate hidden state
        if h0 is None:
            h = torch.zeros(hs.shape[0], self.hidden_units)
            c = torch.zeros(hs.shape[0], self.hidden_units)
        else:
            (h, c) = h0

        # RNN cells
        for i in range(self.n_hidden_layers):
            h = self.rnns[i](hs, h)
            if train:
                hs = self.do(h)
            else:
                hs = h
        y = self.out(hs)
        # return y, (h,c)

        return y


class RNNRegressor(NeuralNetRegressor, TorchSerializationMixin):
    """Basic RNN/LSTM/GRU model.

    Args:
        input_units (int): Number of input units
        output_units (int): Number of output units
        hidden_units (int): Number of hidden units
        n_hidden_layers (int): Number of hidden layers
        output_nonlin (torch.nn.Module instance or None (default=nn.Linear)):
            Non-linearity to apply after last layer, if any.
        dropout (float): Dropout
        squeeze_output (bool): default=False
            Whether to squeeze output. Skorch requirement.
        cell_type (string): default="RNN"
        **kwargs: Arbitrary keyword arguments
    """

    def __init__(
        self,
        input_units,
        output_units,
        hidden_units,
        n_hidden_layers=1,
        dropout=0.2,
        output_nonlin=nn.Linear,
        squeeze_output=False,
        cell_type="RNN",
        **kwargs,
    ):
        super(RNNRegressor, self).__init__(
            module=_RNNModel,
            module__input_units=input_units,
            module__output_units=output_units,
            module__hidden_units=hidden_units,
            module__n_hidden_layers=n_hidden_layers,
            module__dropout=dropout,
            module__output_nonlin=output_nonlin,
            module__squeeze_output=squeeze_output,
            module__cell_type=cell_type,
            train_split=None,
            max_epochs=1,
            batch_size=20,
            **kwargs,
        )



import river.feature_selection.k_best as k_best
import river.feature_selection.random as random
import river.feature_selection.variance as variance
from sail.transformers.river.base import BaseRiverTransformer
from river import stats

__all__ = ["PoissonInclusion", "SelectKBest", "VarianceThreshold"]


class SelectKBest(BaseRiverTransformer):
    def __init__(self, similarity: stats.Bivariate, k=10):
        super(SelectKBest, self).__init__(
            river_estimator=k_best.SelectKBest(similarity, k)
        )


class PoissonInclusion(BaseRiverTransformer):
    def __init__(self, p: float, seed: int = None):
        super(PoissonInclusion, self).__init__(
            river_estimator=random.PoissonInclusion(p, seed)
        )


class VarianceThreshold(BaseRiverTransformer):
    def __init__(self, threshold=0, min_samples=2):
        super(VarianceThreshold, self).__init__(
            river_estimator=variance.VarianceThreshold(threshold, min_samples)
        )



"""
Keras wrapper for The Online Sequential Extreme Learning Machine (OSELM).
"""

import tensorflow as tf
from scikeras.wrappers import KerasRegressor
from tensorflow.python.keras.engine import data_adapter

from sail.models.keras.base import KerasSerializationMixin


class _Model(tf.keras.Model):
    def __init__(
        self,
        num_hidden_nodes: int = 100,
        hidden_layer_activation: str = "sigmoid",
        prediction_window_size: int = 1,
        forgetting_factor: float = 0.9,
    ):
        super(_Model, self).__init__()
        self.prediction_window_size = prediction_window_size
        self.forgetting_factor = forgetting_factor
        self.num_hidden_nodes = num_hidden_nodes
        self.num_output_nodes = self.prediction_window_size
        self.hidden_layer_activation = hidden_layer_activation

    def get_config(self):
        """
        Returns the config of the layer.

        A layer config is a Python dictionary (serializable) containing the configuration of a layer. The same layer can be reinstantiated later (without its trained weights) from this configuration.
        """
        return {
            "num_hidden_nodes": self.num_hidden_nodes,
            "forgetting_factor": self.forgetting_factor,
            "prediction_window_size": self.prediction_window_size,
            "hidden_layer_activation": self.hidden_layer_activation,
        }

    def build(self, input_shape):
        """
        Builds the model and operations needed for training.

        Args:
            input_shape: Single tuple, TensorShape, or list/dict of shapes, where shapes are tuples, integers, or TensorShapes.
        """

        n_input_nodes = input_shape[1]

        # hidden layer
        self.hidden_layer = tf.keras.layers.Dense(
            units=self.num_hidden_nodes,
            input_shape=(n_input_nodes,),
            kernel_initializer=tf.keras.initializers.RandomUniform(
                minval=-1, maxval=1
            ),
            use_bias=True,
            bias_initializer=tf.keras.initializers.RandomUniform(
                minval=-1, maxval=1
            ),
            activation=self.hidden_layer_activation,
            dtype=tf.float32,
            name="hidden_layer",
        )

        self.__p = tf.Variable(
            tf.multiply(
                tf.ones(shape=[self.num_hidden_nodes, self.num_hidden_nodes]),
                0.1,
            ),
            dtype=tf.float32,
            name="p",
        )

        # hidden to output layer connection
        self.__beta = tf.Variable(
            tf.zeros(shape=[self.num_hidden_nodes, self.num_output_nodes]),
            dtype=tf.float32,
            name="beta",
        )

    @tf.autograph.experimental.do_not_convert
    def call(self, inputs, training=None):
        """
        Calls the model on new inputs and returns the outputs as tensors.

        Args:
            inputs: Input tensor, or dict/list/tuple of input tensors.
            training: Boolean or boolean scalar tensor, indicating whether to run
              the `Network` in training mode or inference mode.
            mask: A mask or list of masks. A mask can be
                either a tensor or None (no mask).

        Returns:
            A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        out = self.hidden_layer(inputs)
        if training is not None:
            out = tf.matmul(out, self.__beta)

        return out

    def test_step(self, data):
        """
        Custom logic for one test step.

        Args:
            data: A nested structure of `Tensor`s.

        Returns:
            A `dict` containing values of the `Model`'s metrics are returned.
        """

        # Unpack the data
        data = data_adapter.expand_1d(data)
        (
            features,
            targets,
            _,
        ) = data_adapter.unpack_x_y_sample_weight(data)

        # Compute predictions
        predictions = self(features, training=False)

        # compute loss value
        if self.compiled_loss:
            self.compiled_loss(targets, predictions)

        # Update the metrics.
        if self.compiled_metrics:
            self.compiled_metrics.update_state(targets, predictions)

        return {m.name: m.result() for m in self.metrics}

    def train_step(self, data):
        """
        Custom logic for one training step.

        Args:
            data: A nested structure of `Tensor`s.

        Returns:
            A `dict` containing values of the `Model`'s metrics are returned.
        """

        data = data_adapter.expand_1d(data)

        # OS_ELM being SLFN (single hidden layer feedforward neural network), does not use a loss fn and hence sample_weights are ignored.
        features, targets, _ = data_adapter.unpack_x_y_sample_weight(data)

        H = self(features, training=None)

        ffi = 1.0 / self.forgetting_factor
        HT = tf.transpose(H)
        batch_size = tf.shape(H)[0]
        I = tf.eye(batch_size)
        Hp = tf.matmul(H, ffi * self.__p)
        HpHT = tf.matmul(Hp, HT)
        temp = tf.linalg.pinv(I + HpHT)
        pHT = tf.matmul(ffi * self.__p, HT)
        self.__p.assign(
            tf.subtract(ffi * self.__p, tf.matmul(tf.matmul(pHT, temp), Hp))
        )
        pHT = tf.matmul(self.__p, HT)
        Hbeta = tf.matmul(H, self.__beta)
        self.__beta.assign(
            self.__beta + tf.matmul(pHT, tf.subtract(targets, Hbeta))
        )

        # use __beta to make final prediction
        predictions = tf.matmul(H, self.__beta)

        # compute loss value
        if self.compiled_loss:
            self.compiled_loss(targets, predictions)

        # Update the metrics.
        if self.compiled_metrics:
            self.compiled_metrics.update_state(targets, predictions)

        return {m.name: m.result() for m in self.metrics}


class OSELM(KerasRegressor, KerasSerializationMixin):
    """
    Keras wrapper for The Online Sequential Extreme Learning Machine (OSELM).

    The Online Sequential Extreme Learning Machine (OSELM) is an online sequential learning algorithm for single hidden layer feed forward neural networks that learns the train data one-by-one or chunk-by-chunk without retraining all the historic data. It gives better generalization performance at very fast learning speed.

    OSELM being SLFN (single hidden layer feedforward neural network), does not use a loss fn to calculate gradients. It has no other control parameters for users to mannually tuning.

    Parameters
    ----------

        optimizer : Union[str, tf.keras.optimizers.Optimizer, Type[tf.keras.optimizers.Optimizer]], default "adam"
            This can be a string for Keras' built in optimizersan instance of tf.keras.optimizers.Optimizer or a class inheriting from tf.keras.optimizers.Optimizer. Only strings and classes support parameter routing.

        loss : Union[Union[str, tf.keras.losses.Loss, Type[tf.keras.losses.Loss], Callable], None], default="mse"
            The loss function to use for training. This can be a string for Keras' built in losses, an instance of tf.keras.losses.Loss or a class inheriting from tf.keras.losses.Loss .

        metrics : List[str], default=None
            list of metrics to be evaluated.

        num_hidden_nodes : int, default=100
            number of neurons to use in the hidden layer.

        hidden_layer_activation : str, default=sigmoid
            Activation function to apply on the hidden layer - can use any tf.keras.activation function name.

        prediction_window_size: int, default=1
            number of timeseries steps to predict in the future.

        forgetting_factor: float, default=0.9
            The forgetting factor can be set in order to forget the old observations quickly and track the new data tightly.

        epochs: int, default=1
            Number of training steps.

        verbose: int default=0
            To display/supress logs.
    """

    def __init__(
        self,
        loss="mse",
        optimizer=tf.keras.optimizers.Adam,
        metrics=None,
        epochs=1,
        verbose=0,
        num_hidden_nodes=25,
        hidden_layer_activation=tf.nn.sigmoid,
        prediction_window_size=1,
        forgetting_factor=0.5,
        **kwargs,
    ) -> None:
        super(OSELM, self).__init__(
            _Model(
                num_hidden_nodes,
                hidden_layer_activation,
                prediction_window_size,
                forgetting_factor,
            ),
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            epochs=epochs,
            verbose=verbose,
            **kwargs,
        )
        self.prediction_window_size = prediction_window_size

    def _ensure_compiled_model(self) -> None:
        super()._ensure_compiled_model()
        self.model_.outputs = [1] * self.prediction_window_size



import copy
from river.compat.river_to_sklearn import SKLEARN_INPUT_X_PARAMS, SKLEARN_INPUT_Y_PARAMS
import numpy as np
from river import base
from river.compat import River2SKLTransformer, river_to_sklearn
from sklearn import utils
from sail.common.mixin import RiverAttributeMixin


class BaseRiverTransformer(RiverAttributeMixin, River2SKLTransformer):
    def __init__(self, *args, **Kwargs):
        super(BaseRiverTransformer, self).__init__(*args, **Kwargs)
        self.validation_params = {"cast_to_ndarray": True}

    def _partial_fit(self, X, y=None):
        # Check the inputs
        first_call = not hasattr(self, "n_features_in_")

        if y is None:
            X = self._validate_data(
                X,
                **self.validation_params,
                reset=first_call,
                **SKLEARN_INPUT_X_PARAMS,
            )
        else:
            X, y = self._validate_data(
                X,
                y,
                **self.validation_params,
                reset=first_call,
                **SKLEARN_INPUT_X_PARAMS,
                **SKLEARN_INPUT_Y_PARAMS,
            )

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to deep copy the provided estimator in order to respect this convention
        if not hasattr(self, "instance_"):
            self.instance_ = copy.deepcopy(self.river_estimator)

        # Call learn_one for each observation
        if isinstance(self.instance_, base.SupervisedTransformer):
            for x, yi in river_to_sklearn.STREAM_METHODS[type(X)](X, y):
                self.instance_.learn_one(x, yi)
        else:
            for x, _ in river_to_sklearn.STREAM_METHODS[type(X)](X):
                self.instance_.learn_one(x)

        return self

    def transform(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters
        ----------
        X
            array-like of shape (n_samples, n_features)

        Returns
        -------
        Transformed output.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, attributes="instance_")

        # Check the input
        X = self._validate_data(
            X, **self.validation_params, reset=False, **SKLEARN_INPUT_X_PARAMS
        )

        # Call predict_proba_one for each observation
        X_trans = [None] * len(X)
        for i, (x, _) in enumerate(river_to_sklearn.STREAM_METHODS[type(X)](X)):
            X_trans[i] = list(self.instance_.transform_one(x).values())

        return np.asarray(X_trans)

    def partial_fit_transform(self, X, y=None, **fit_params):
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            return self.partial_fit(X, **fit_params).transform(X)
        else:
            # fit method of arity 2 (supervised transformation)
            return self.partial_fit(X, y, **fit_params).transform(X)



import os

from sail.utils.logging import configure_logger
from sail.utils.serialization import load_obj, save_obj

LOGGER = configure_logger(logger_name="SAILModel")


class SAILModel:
    def fit(self):
        raise NotImplementedError

    def partial_fit(self):
        raise NotImplementedError

    def predict(self):
        raise NotImplementedError

    def score(self):
        raise NotImplementedError

    def save_model(self, model_folder):
        """Saves the sail model to pickle format.

        Args:
            model_folder: String, PathLike, path to SavedModel.
        """
        model_path = os.path.join(model_folder, "model")
        save_obj(
            self,
            model_path,
            str(self.__class__.__name__),
            serialize_type="joblib",
        )
        LOGGER.info("Model saved successfully.")

    def load_model(self, model_folder):
        """Load the sail model from the mentioned folder location

        Args:
            model_folder: One of the following:
                - String or `pathlib.Path` object, path to the saved model

        Returns:
            BaseModel: Model
        """
        model_path = os.path.join(model_folder, "model")
        model = load_obj(
            model_path,
            str(self.__class__.__name__),
            serialize_type="joblib",
        )
        LOGGER.info("Model loaded successfully.")
        return model



from sail.utils.logging import configure_logger
import os
from pathlib import Path

LOGGER = configure_logger(logger_name="PyTorch")


class TorchSerializationMixin:
    def save(self, model_folder):
        """
        Saves the module's parameters, history, and optimizer.

        Args:
            model_folder: String, PathLike, path to model.
        """
        metadata_path = os.path.join(model_folder, "metadata")
        Path(metadata_path).mkdir(parents=True, exist_ok=True)
        self.save_params(  # type: ignore
            f_params=metadata_path + "/model.pkl",
            f_optimizer=metadata_path + "/opt.pkl",
            f_history=metadata_path + "/history.json",
        )
        LOGGER.info("Model saved successfully.")

    def load(self, model_folder):
        """
        Loads the the module's parameters, history, and optimizer.

        Args:
            model_folder: One of the following:
                - String or `pathlib.Path` object, path to the saved model
        """
        metadata_path = os.path.join(model_folder, "metadata")
        if os.path.exists(metadata_path):
            self.load_params(  # type: ignore
                f_params=metadata_path + "/model.pkl",
                f_optimizer=metadata_path + "/opt.pkl",
                f_history=metadata_path + "/history.json",
            )



import numpy as np
from pandas import DataFrame


def generate_features_and_targets(time_series: DataFrame, input_column: str):
    """Generate features and targets from a timeseries with a lag of 1.

    Parameters
    ----------
    ts_df: Pandas Dataframe
        time series dataset as a Pandas Dataframe.
    input_column: str
        name of the input column in the dataset.

    Returns
    -------
    ndarray
        Numpy array containing features values.
    ndarray
        Numpy array containing target values. Lag by 1.
    """

    num_rows = len(time_series)
    features = time_series[input_column].astype(float)[0 : num_rows - 1]
    targets = time_series[input_column].astype(float)[1:num_rows]

    return np.array(features).reshape(-1, 1), np.array(targets).reshape(-1, 1)



from opentelemetry import trace as oltp_trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter,
    _append_trace_path,
)
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanExporter
from opentelemetry.trace.propagation import set_span_in_context


def trace_with_action(span_name, current_span=False):
    def trace_span(func):
        def wrapper(*args, **kwargs):
            base_class = args[0]

            if base_class.tracer is not None:
                verbosity = base_class.verbosity
                if current_span:
                    with base_class.tracer.trace_as_current_span(
                        span_name=f"Epoch-{verbosity.current_epoch_n}-{span_name}",
                        verbose=verbosity.get(),
                    ):
                        return func(*args, **kwargs)
                else:
                    with base_class.tracer.trace(
                        span_name=f"Epoch-{verbosity.current_epoch_n}-{span_name}",
                        verbose=verbosity.get(),
                    ):
                        return func(*args, **kwargs)
            else:
                return func(*args, **kwargs)

        return wrapper

    return trace_span


def trace(span_name, current_span=False):
    def trace_span(func):
        def wrapper(*args, **kwargs):
            base_class = args[0]

            if base_class.tracer is not None:
                if current_span:
                    with base_class.tracer.trace_as_current_span(span_name=span_name):
                        return func(*args, **kwargs)
                else:
                    with base_class.tracer.trace(span_name=span_name):
                        return func(*args, **kwargs)
            else:
                return func(*args, **kwargs)

        return wrapper

    return trace_span


class TracingClient:
    def __init__(
        self,
        service_name: str,
        otlp_endpoint: str = None,
        span_exporter: SpanExporter = None,
    ) -> None:
        self.tracer_name = service_name
        self.span_exporter = span_exporter

        provider = TracerProvider(
            resource=Resource.create({SERVICE_NAME: service_name})
        )

        if otlp_endpoint is not None:
            span_exporter = OTLPSpanExporter(endpoint=_append_trace_path(otlp_endpoint))
        elif span_exporter is None:
            span_exporter = OTLPSpanExporter()

        provider.add_span_processor(BatchSpanProcessor(span_exporter))
        oltp_trace.set_tracer_provider(provider)

        self._tracer = oltp_trace.get_tracer(service_name)

    def trace_as_current_span(
        self,
        span_name,
        kind=oltp_trace.SpanKind.INTERNAL,
        context=None,
        verbose=1,
    ):
        if verbose == 1:
            return self._tracer.start_as_current_span(
                span_name, context=context, kind=kind
            )
        else:
            return DummySpan()

    def trace(
        self,
        span_name,
        kind=oltp_trace.SpanKind.INTERNAL,
        context=None,
        verbose=1,
    ):
        if verbose == 1:
            return self._tracer.start_span(span_name, context=context, kind=kind)
        else:
            return DummySpan()

    def set_attribute(self, name, value):
        oltp_trace.get_current_span().set_attribute(name, value)


class DummySpan:
    def __enter__(self):
        return self

    def __exit__(self, *args):
        return self



import copy
import importlib
import os
import shutil
from typing import Literal, Type, Union

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils import check_array, check_X_y

from sail.common.decorators import validate_X_y
from sail.common.helper import VerboseManager
from sail.drift_detection.drift_detector import SAILDriftDetector
from sail.models.auto_ml.base_strategy import PipelineActionType, PipelineStrategy
from sail.models.auto_ml.pipeline_strategy import DetectAndIncrement
from sail.models.auto_ml.tune import SAILTuneGridSearchCV, SAILTuneSearchCV
from sail.models.base import SAILModel
from sail.pipeline import SAILPipeline
from sail.telemetry import TracingClient
from sail.utils.logging import configure_logger
from sail.utils.serialization import load_obj, save_obj

LOGGER = configure_logger(logger_name="SAILAutoPipeline")


class SAILAutoPipeline(SAILModel, BaseEstimator):
    def __init__(
        self,
        pipeline: SAILPipeline,
        pipeline_params_grid: dict,
        search_method: Union[None, str] = None,
        search_method_params: dict = None,
        search_data_size: int = 1000,
        incremental_training: bool = False,
        drift_detector: Union[None, SAILDriftDetector] = None,
        pipeline_strategy: Union[None, str] = None,
        verbosity_level: Literal[0, 1] | None = 1,
        verbosity_interval: int | None = None,
        tensorboard_log_dir: str = None,
        tracer: TracingClient = None,
    ) -> None:
        self.pipeline = pipeline
        self.pipeline_params_grid = pipeline_params_grid
        self.search_method_params = search_method_params
        self.search_data_size = search_data_size
        self.incremental_training = incremental_training
        self.drift_detector = drift_detector
        if drift_detector is None:
            self.drift_detector = SAILDriftDetector()
        self.verbosity_level = verbosity_level
        self.verbosity_interval = verbosity_interval
        self.tensorboard_log_dir = tensorboard_log_dir
        self.tracer = tracer

        self.search_method = self._resolve_search_method(
            search_method, search_method_params
        )
        # validate and create verbosity manager
        self.verbosity = self._resolve_verbosity(verbosity_level, verbosity_interval)
        self.pipeline_strategy = self._resolve_pipeline_strategy(pipeline_strategy)

    @property
    def best_pipeline(self) -> SAILPipeline:
        if hasattr(self.pipeline_strategy, "_best_pipeline"):
            return self.pipeline_strategy._best_pipeline
        return None

    @property
    def progressive_score(self) -> float:
        if self.check_is_fitted("progressive_score"):
            return self.best_pipeline.get_progressive_score

    @property
    def cv_results(self):
        self.check_is_fitted("cv_results()")
        if hasattr("fit_result", self):
            return self.pipeline_strategy._fit_result.cv_results_

    def check_is_fitted(self, func) -> None:
        pipeline_action = self.pipeline_strategy.pipeline_actions.current_action_node
        if pipeline_action.action == PipelineActionType.DATA_COLLECTION:
            LOGGER.warning(
                f"Ignoring...{func}, since the best pipeline is not available yet. Keep calling 'train' with new data to get the best pipeline."
            )
            fitted = False
        elif (
            pipeline_action.previous.action == PipelineActionType.SCORE_AND_DETECT_DRIFT
            and hasattr(self.pipeline_strategy, "_best_pipeline")
        ):
            additional_msg = (
                " Please note that input data of length {self.search_data_size} is required to begin new parameter search."
                if self.pipeline_strategy
                in ["DetectAndWarmStart", "DetectAndRestart", "PrequentialTraining"]
                else ""
            )
            LOGGER.warning(
                f"The current best pipeline is STALE. Pipeline becomes stale when data drift occurs. You can call 'train' with fresh data to get the best pipeline."
                + additional_msg
            )
            fitted = True
        elif not hasattr(self.pipeline_strategy, "_best_pipeline"):
            raise Exception(
                f"The current instance is not fitted yet. Call 'train' with appropriate arguments before using {func}."
            )
        elif hasattr(self.pipeline_strategy, "_best_pipeline"):
            fitted = True

        return fitted

    def _validate_X_y(self, X, y=None):
        if len(X.shape) == 1:
            X = X.reshape((1, X.shape[0]))

        X_new = X.copy()
        datetime_cols = list(X_new.select_dtypes(include="datetime64[ns]"))
        X_new = X_new.drop(datetime_cols, axis=1)

        if y is None:
            _ = check_array(
                X_new,
                dtype=None,
                input_name="X",
            )
            return X
        else:
            _, y = check_X_y(X_new, y, dtype=None)
            return X, y

    def _resolve_verbosity(self, verbosity, verbosity_interval):
        if verbosity is None:
            return VerboseManager(
                "SAILPipeline",
                0,
                verbosity_interval,
            )
        elif isinstance(verbosity, int):
            return VerboseManager(
                "SAILPipeline",
                verbosity,
                verbosity_interval,
            )

    def _resolve_search_method(self, search_method, search_method_params):
        if search_method is None:
            _search_class = SAILTuneGridSearchCV
        elif Type[search_method] in [
            Type[SAILTuneGridSearchCV],
            Type[SAILTuneSearchCV],
        ]:
            _search_class = search_method
        elif isinstance(search_method, str):
            module = importlib.import_module("sail.models.auto_ml.tune")
            try:
                _search_class = getattr(module, search_method)
            except AttributeError:
                raise Exception(
                    f"Method '{search_method}' is not available. search_method must be from [SAILTuneGridSearchCV, SAILTuneSearchCV] from the module sail.models.auto_ml.tune. Set `None` to use the default."
                )
        else:
            raise TypeError(
                "`search_method` must be None or an instance or str from "
                f"[SAILTuneGridSearchCV, SAILTuneSearchCV] from sail.models.auto_ml.tune. Got {search_method.__module__}.{search_method.__class__.__qualname__}. Set `None` to use the default."
            )

        if search_method_params is None:
            search_method_params = _search_class.default_search_params
        else:
            # update params from the default ones if missing any.
            search_method_params = {
                **_search_class.default_search_params,
                **search_method_params,
            }

        return _search_class(
            estimator=self.pipeline,
            param_grid=self.pipeline_params_grid,
            **search_method_params,
        )

    def _resolve_pipeline_strategy(self, pipeline_strategy):
        pipeline_strategy_class = None
        if pipeline_strategy is None:
            pipeline_strategy_class = DetectAndIncrement
        elif isinstance(pipeline_strategy, str):
            if pipeline_strategy in PipelineStrategy.defined_stategies:
                module = importlib.import_module(
                    "sail.models.auto_ml.pipeline_strategy"
                )
                pipeline_strategy_class = getattr(module, pipeline_strategy)
            else:
                raise ValueError(
                    "{} is not a defined pipeline strategy. "
                    "Please select from the list of available strategies: {}".format(
                        pipeline_strategy, PipelineStrategy.defined_stategies
                    )
                )
        else:
            raise TypeError(
                "`pipeline_strategy` must be a None, str, "
                f"or an instance of PipelineStrategy. Got {pipeline_strategy.__module__}.{pipeline_strategy.__class__.__qualname__}. Set `None` to use the default."
            )

        return pipeline_strategy_class(
            search_method=self.search_method,
            search_data_size=self.search_data_size,
            drift_detector=self.drift_detector,
            verbosity=self.verbosity,
            incremental_training=self.incremental_training,
            tensorboard_log_dir=self.tensorboard_log_dir,
            tracer=self.tracer,
        )

    @validate_X_y
    def train(self, X, y=None, **fit_params):
        self.pipeline_strategy.next(X, y, **fit_params)

    def predict(self, X, **predict_params):
        if self.check_is_fitted("predict()"):
            X = self._validate_X_y(X)
            return self.best_pipeline.predict(X, **predict_params)

    def score(self, X, y=None, sample_weight=1.0) -> float:
        if self.check_is_fitted("score()"):
            return self.best_pipeline.score(X, y, sample_weight)

    def save_model(self, model_folder, overwrite=True) -> str:
        """
        Parameters:
        -----------
        model_folder: str
            location to save the model
        overwrite: bool (False)
            if True and model_folder already exists, it will be delted and recreated

        Returns
        -------
        saved_location: str
        """
        save_location = os.path.join(model_folder, "sail_auto_pipeline")
        if not overwrite and os.path.exists(save_location):
            raise Exception(
                f"{save_location} already exists, specify overwrite=True to replace contents"
            )
        else:
            # delete old folder to avoid unwanted file overlaps
            if os.path.exists(save_location) and os.path.isdir(save_location):
                shutil.rmtree(save_location)

            LOGGER.debug(f"making directory tree {save_location}")
            os.makedirs(save_location, exist_ok=True)
            if not os.path.exists(save_location):
                raise Exception(f"target directory {save_location} can not be created!")

        # -------------------------------------------
        # Get all params
        # -------------------------------------------
        params = self.get_params(deep=False)

        # -------------------------------------------
        # save search_method params
        # -------------------------------------------
        obj = []
        search_method = params.pop("search_method")
        if hasattr(search_method, "_best_configurations"):
            obj = search_method._best_configurations
        save_obj(
            obj,
            location=os.path.join(save_location, "search_method"),
            file_name="best_configurations",
        )

        # -------------------------------------------
        # save pipeline_strategy state
        # -------------------------------------------
        pipeline_strategy = params.pop("pipeline_strategy")
        save_obj(
            {"current_action": pipeline_strategy.get_current_action()},
            location=os.path.join(save_location, "pipeline_strategy"),
            file_name="state",
        )

        # -------------------------------------------
        # save pipeline_strategy verbosity state
        # -------------------------------------------
        save_obj(
            obj=self.verbosity.get_state(),
            location=os.path.join(save_location, "pipeline_strategy"),
            file_name="verbosity_state",
            serialize_type="json",
        )

        # -------------------------------------------
        # save tensorboard writer state if enabled
        # -------------------------------------------
        if self.tensorboard_log_dir:
            save_obj(
                obj=self.pipeline_strategy.writer.get_state(),
                location=os.path.join(save_location, "pipeline_strategy"),
                file_name="tensorboard_dir",
                serialize_type="json",
            )

        # -------------------------------------------
        # save data already collected for auto ml tuning
        # -------------------------------------------
        if hasattr(pipeline_strategy, "_input_X"):
            np.savez(
                os.path.join(save_location, "pipeline_strategy", "data"),
                input_X=pipeline_strategy._input_X,
                input_y=pipeline_strategy._input_y,
            )

        # -------------------------------------------
        # save fit results and best pipeline
        # -------------------------------------------
        if hasattr(pipeline_strategy, "_best_pipeline"):
            # save_obj(
            #     pipeline_strategy._fit_result,
            #     location=os.path.join(save_location, "pipeline_strategy"),
            #     file_name="fit_result",
            # )
            pipeline_strategy._best_pipeline.save(
                os.path.join(save_location, "pipeline_strategy"), name="best_pipeline"
            )

        # -------------------------------------------
        # save rest of the params
        # -------------------------------------------
        save_obj(
            obj=params,
            location=save_location,
            file_name="params",
        )

        return save_location

    @classmethod
    def load_model(cls, model_folder):
        load_location = os.path.join(model_folder, "sail_auto_pipeline")

        # -------------------------------------------
        # Load params
        # -------------------------------------------
        params = load_obj(location=load_location, file_name="params")

        # -------------------------------------------
        # create SAILAutoPipeline
        # -------------------------------------------
        sail_auto_pipeline = SAILAutoPipeline(**params)

        # -------------------------------------------
        # Load best_configurations if present
        # -------------------------------------------
        best_configurations = load_obj(
            location=os.path.join(load_location, "search_method"),
            file_name="best_configurations",
        )
        if len(best_configurations) > 0:
            sail_auto_pipeline.search_method._best_configurations = best_configurations

        # -------------------------------------------
        # Load existing pipeline_strategy state
        # -------------------------------------------
        state = load_obj(
            location=os.path.join(load_location, "pipeline_strategy"),
            file_name="state",
        )
        sail_auto_pipeline.pipeline_strategy.set_current_action(state["current_action"])

        # -------------------------------------------
        # load verbosity state
        # -------------------------------------------
        state = load_obj(
            location=os.path.join(load_location, "pipeline_strategy"),
            file_name="verbosity_state",
            serialize_type="json",
        )
        sail_auto_pipeline.verbosity.set_state(state)

        # -------------------------------------------
        # load tensorboard writer state if enabled
        # -------------------------------------------
        if sail_auto_pipeline.tensorboard_log_dir:
            state = load_obj(
                location=os.path.join(load_location, "pipeline_strategy"),
                file_name="tensorboard_dir",
                serialize_type="json",
            )
            sail_auto_pipeline.pipeline_strategy.writer.set_state(state)

        # -------------------------------------------
        # Load data already collected for auto ml tuning
        # -------------------------------------------
        if os.path.exists(os.path.join(load_location, "pipeline_strategy", "data.npz")):
            data = np.load(os.path.join(load_location, "pipeline_strategy", "data.npz"))
            sail_auto_pipeline.pipeline_strategy._input_X = data["input_X"]
            sail_auto_pipeline.pipeline_strategy._input_y = data["input_y"]
            data.close()

        # -------------------------------------------
        # Load fit results and best pipeline
        # -------------------------------------------
        try:
            # fit_result = load_obj(
            #     location=os.path.join(load_location, "pipeline_strategy"),
            #     file_name="fit_result",
            # )
            # sail_auto_pipeline.pipeline_strategy._fit_result = fit_result
            best_pipeline = SAILPipeline.load(
                os.path.join(load_location, "pipeline_strategy"), name="best_pipeline"
            )
            sail_auto_pipeline.pipeline_strategy._best_pipeline = best_pipeline

            # replace verbosity instance of the best pipeline from the parent.
            sail_auto_pipeline.pipeline_strategy._best_pipeline.verbosity = (
                sail_auto_pipeline.verbosity
            )

        except Exception as e:
            print(e)

        return sail_auto_pipeline




import torch
from torch import nn
import torch.nn.functional as F

from sklearn.datasets import make_classification
from sklearn.preprocessing import MinMaxScaler
import numpy as np


from typing import cast, Union, List

from skorch import NeuralNetClassifier

'''
Define conv1d with same padding
'''



class Conv1dSamePadding(nn.Conv1d):
    """Represents the "Same" padding functionality from Tensorflow.
    See: https://github.com/pytorch/pytorch/issues/3867
    Note that the padding argument in the initializer doesn't do anything now
    """
    def forward(self, input):
        return conv1d_same_padding(input, self.weight, self.bias, self.stride,
                                   self.dilation, self.groups)


def conv1d_same_padding(input, weight, bias, stride, dilation, groups):
    # stride and dilation are expected to be tuples.
    kernel, dilation, stride = weight.size(2), dilation[0], stride[0]
    l_out = l_in = input.size(2)
    padding = (((l_out - 1) * stride) - l_in + (dilation * (kernel - 1)) + 1)
    if padding % 2 != 0:
        input = F.pad(input, [0, 1])

    return F.conv1d(input=input, weight=weight, bias=bias, stride=stride,
                    padding=padding // 2,
                    dilation=dilation, groups=groups)


class ConvBlock(nn.Module):

    def __init__(self, in_channels: int, out_channels: int, kernel_size: int,
                 stride: int) -> None:
        super().__init__()

        self.layers = nn.Sequential(
            Conv1dSamePadding(in_channels=in_channels,
                              out_channels=out_channels,
                              kernel_size=kernel_size,
                              stride=stride),
            nn.BatchNorm1d(num_features=out_channels),
            nn.ReLU(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore

        return self.layers(x)


class _InceptionModel(nn.Module):
    """A PyTorch implementation of the InceptionTime model.
    From https://arxiv.org/abs/1909.04939
    Attributes
    ----------
    num_blocks:
        The number of inception blocks to use. One inception block consists
        of 3 convolutional layers, (optionally) a bottleneck and (optionally) a residual
        connector
    in_channels:
        The number of input channels (i.e. input.shape[-1])
    out_channels:
        The number of "hidden channels" to use. Can be a list (for each block) or an
        int, in which case the same value will be applied to each block
    bottleneck_channels:
        The number of channels to use for the bottleneck. Can be list or int. If 0, no
        bottleneck is applied
    kernel_sizes:
        The size of the kernels to use for each inception block. Within each block, each
        of the 3 convolutional layers will have kernel size
        `[kernel_size // (2 ** i) for i in range(3)]`
    num_pred_classes:
        The number of output classes
    """

    def __init__(self, num_blocks: int, in_channels: int, out_channels: Union[List[int], int],
                 bottleneck_channels: Union[List[int], int], kernel_sizes: Union[List[int], int],
                 use_residuals: Union[List[bool], bool, str] = 'default',
                 num_pred_classes: int = 1
                 ) -> None:
        super().__init__()

        # for easier saving and loading
        self.input_args = {
            'num_blocks': num_blocks,
            'in_channels': in_channels,
            'out_channels': out_channels,
            'bottleneck_channels': bottleneck_channels,
            'kernel_sizes': kernel_sizes,
            'use_residuals': use_residuals,
            'num_pred_classes': num_pred_classes
        }

        channels = [in_channels] + cast(List[int], self._expand_to_blocks(out_channels,
                                                                          num_blocks))
        bottleneck_channels = cast(List[int], self._expand_to_blocks(bottleneck_channels,
                                                                     num_blocks))
        kernel_sizes = cast(List[int], self._expand_to_blocks(kernel_sizes, num_blocks))
        if use_residuals == 'default':
            use_residuals = [True if i % 3 == 2 else False for i in range(num_blocks)]
        use_residuals = cast(List[bool], self._expand_to_blocks(
            cast(Union[bool, List[bool]], use_residuals), num_blocks)
        )

        self.blocks = nn.Sequential(*[
            _InceptionBlock(in_channels=channels[i], out_channels=channels[i + 1],
                           residual=use_residuals[i], bottleneck_channels=bottleneck_channels[i],
                           kernel_size=kernel_sizes[i]) for i in range(num_blocks)
        ])

        # a global average pooling (i.e. mean of the time dimension) is why
        # in_features=channels[-1]
        self.linear = nn.Linear(in_features=channels[-1], out_features=num_pred_classes)

    @staticmethod
    def _expand_to_blocks(value: Union[int, bool, List[int], List[bool]],
                          num_blocks: int) -> Union[List[int], List[bool]]:
        if isinstance(value, list):
            assert len(value) == num_blocks, \
                f'Length of inputs lists must be the same as num blocks, ' \
                f'expected length {num_blocks}, got {len(value)}'
        else:
            value = [value] * num_blocks
        return value

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        x = self.blocks(x).mean(dim=-1)  # the mean is the global average pooling
        return self.linear(x)


class _InceptionBlock(nn.Module):
    """An inception block consists of an (optional) bottleneck, followed
    by 3 conv1d layers. Optionally residual
    """

    def __init__(self, in_channels: int, out_channels: int,
                 residual: bool, stride: int = 1, bottleneck_channels: int = 32,
                 kernel_size: int = 41) -> None:
        assert kernel_size > 3, "Kernel size must be strictly greater than 3"
        super().__init__()

        self.use_bottleneck = bottleneck_channels > 0
        if self.use_bottleneck:
            self.bottleneck = Conv1dSamePadding(in_channels, bottleneck_channels,
                                                kernel_size=1, bias=False)
        kernel_size_s = [kernel_size // (2 ** i) for i in range(3)]
        start_channels = bottleneck_channels if self.use_bottleneck else in_channels
        channels = [start_channels] + [out_channels] * 3
        self.conv_layers = nn.Sequential(*[
            Conv1dSamePadding(in_channels=channels[i], out_channels=channels[i + 1],
                              kernel_size=kernel_size_s[i], stride=stride, bias=False)
            for i in range(len(kernel_size_s))
        ])

        self.batchnorm = nn.BatchNorm1d(num_features=channels[-1])
        self.relu = nn.ReLU()

        self.use_residual = residual
        if residual:
            self.residual = nn.Sequential(*[
                Conv1dSamePadding(in_channels=in_channels, out_channels=out_channels,
                                  kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm1d(out_channels),
                nn.ReLU()
            ])

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        org_x = x
        if self.use_bottleneck:
            x = self.bottleneck(x)
        x = self.conv_layers(x)

        if self.use_residual:
            x = x + self.residual(org_x)
        return x

'''
wrapper scorch class
'''


class InceptionTimeClassifier(NeuralNetClassifier):
  def __init__(self, num_blocks=2, in_channels=1, out_channels=2,
                           bottleneck_channels=2, kernel_sizes=41, use_residuals=True,
                           num_pred_classes=1, learning_rate=0.05, batch_size=1000, criterion = nn.BCEWithLogitsLoss, max_epochs = 50, device='cpu'):
    
    self.inception_model = _InceptionModel(num_blocks=2, in_channels=1, out_channels=2,
                           bottleneck_channels=2, kernel_sizes=41, use_residuals=True,
                           num_pred_classes=1)
    
    super(InceptionTimeClassifier, self).__init__(
            module=self.inception_model,
            max_epochs=max_epochs,
            lr=learning_rate,
            batch_size=batch_size,
            criterion=criterion ,
            # Shuffle training data on each epoch
            iterator_train__shuffle=False,
            device=device,
            )



import os
import json
import dill
import joblib


def save_obj(obj, location, file_name, serialize_type="dill", **kwargs):
    """
    serialize object state
    """
    if not os.path.exists(location):
        os.makedirs(location)

    if serialize_type == "joblib":
        with open(os.path.join(location, file_name + ".joblib"), "wb") as save_file:
            joblib.dump(obj, save_file)
    elif serialize_type == "dill":
        with open(os.path.join(location, file_name + ".pickle"), "wb") as save_file:
            dill.dump(obj, save_file)
    elif serialize_type == "json":
        with open(os.path.join(location, file_name + ".json"), "w") as save_file:
            json_encoder = kwargs["cls"] if "cls" in kwargs else None
            json.dump(obj, save_file, cls=json_encoder)


def load_obj(location, file_name, serialize_type="dill", **kwargs):
    """
    deserialize object state
    """

    if serialize_type == "joblib":
        with open(os.path.join(location, file_name + ".joblib"), "rb") as load_file:
            return joblib.load(load_file)
    elif serialize_type == "dill":
        with open(os.path.join(location, file_name + ".pickle"), "rb") as load_file:
            return dill.load(load_file)
    elif serialize_type == "json":
        with open(os.path.join(location, file_name + ".json"), "r") as save_file:
            json_encoder = kwargs["cls"] if "cls" in kwargs else None
            return json.load(save_file, cls=json_encoder)



import typing

import river.ensemble.bagging as bagging
import river.ensemble.boosting as boosting
import river.ensemble.streaming_random_patches as streaming_random_patches
from river import base
from river.metrics.base import Metric

from sail.models.river.base import SailRiverClassifier, SailRiverRegressor

__all__ = [
    "AdaBoostClassifier",
    "ADWINBaggingClassifier",
    "BaggingClassifier",
    "BaggingRegressor",
    "LeveragingBaggingClassifier",
    "SRPClassifier",
]


class BaggingClassifier(SailRiverClassifier):
    def __init__(self, model: base.Classifier, n_models=10, seed: int = None):
        super(BaggingClassifier, self).__init__(
            river_estimator=bagging.BaggingClassifier(model, n_models, seed)
        )


class BaggingRegressor(SailRiverRegressor):
    def __init__(self, model: base.Regressor, n_models=10, seed: int = None):
        super(BaggingRegressor, self).__init__(
            river_estimator=bagging.BaggingRegressor(model, n_models, seed)
        )


class ADWINBaggingClassifier(SailRiverClassifier):
    def __init__(self, model: base.Classifier, n_models=10, seed: int = None):
        super(ADWINBaggingClassifier, self).__init__(
            river_estimator=bagging.ADWINBaggingClassifier(model, n_models, seed)
        )


class LeveragingBaggingClassifier(SailRiverClassifier):
    def __init__(
        self,
        model: base.Classifier,
        n_models: int = 10,
        w: float = 6,
        adwin_delta: float = 0.002,
        bagging_method: str = "bag",
        seed: int = None,
    ):
        super(LeveragingBaggingClassifier, self).__init__(
            river_estimator=bagging.LeveragingBaggingClassifier(
                model, n_models, w, adwin_delta, bagging_method, seed
            )
        )


class AdaBoostClassifier(SailRiverClassifier):
    def __init__(self, model: base.Classifier, n_models=10, seed: int = None):
        super(AdaBoostClassifier, self).__init__(
            river_estimator=boosting.AdaBoostClassifier(model, n_models, seed)
        )


class SRPClassifier(SailRiverClassifier):
    def __init__(
        self,
        model: base.Estimator = None,
        n_models: int = 10,
        subspace_size: typing.Union[int, float, str] = 0.6,
        training_method: str = "patches",
        lam: float = 6.0,
        drift_detector: base.DriftDetector = None,
        warning_detector: base.DriftDetector = None,
        disable_detector: str = "off",
        disable_weighted_vote: bool = False,
        seed=None,
        metric: Metric = None,
    ):
        super(SRPClassifier, self).__init__(
            river_estimator=streaming_random_patches.SRPClassifier(
                model,
                n_models,
                subspace_size,
                training_method,
                lam,
                drift_detector,
                warning_detector,
                disable_detector,
                disable_weighted_vote,
                seed,
                metric,
            )
        )



import inspect
import typing


class RiverAttributeMixin:
    def __getattr__(self, __name: str) -> typing.Any:
        return getattr(self.river_estimator, __name)

    def __setattr__(self, __name: str, __value: typing.Any) -> None:
        if __name in list(inspect.signature(__class__).parameters):
            setattr(self.river_estimator, __name, __value)
        else:
            super().__setattr__(__name, __value)



from river import base, compose
from sail.transformers.river.base import BaseRiverTransformer
from sklearn.base import ClassNamePrefixFeaturesOutMixin
from sklearn.utils.validation import check_is_fitted
from typing import List


class Select(ClassNamePrefixFeaturesOutMixin, BaseRiverTransformer):
    def __init__(self, keys: List[base.typing.FeatureName]):
        super(Select, self).__init__(river_estimator=compose.Select(*keys))
        self.keys = keys
        self.validation_params["cast_to_ndarray"] = False

    def get_feature_names_out(self, input_features=None):
        """Get output feature names for transformation.

        Parameters
        ----------
        input_features : array-like of str or None, default=None
            Input features.

            - If `input_features` is `None`, then `feature_names_in_` is
              used as feature names in. If `feature_names_in_` is not defined,
              then the following input feature names are generated:
              `["x0", "x1", ..., "x(n_features_in_ - 1)"]`.
            - If `input_features` is an array-like, then `input_features` must
              match `feature_names_in_` if `feature_names_in_` is defined.

        Returns
        -------
        feature_names_out : ndarray of str objects
            Same as input features.
        """
        check_is_fitted(self, "n_features_in_")
        return list(self.instance_.keys)



# https://github.com/Jung-Sanghyun/Incremental-SVM/blob/main/HMISVM.py
# https://github.com/galrettig/PegasosSvm/blob/master/Pegasos_Svm_Explore.ipynb
import numpy as np


class ISVM:
    def __init__(self, lam=0.001, T=10000):
        self.lam = lam
        self.T = T
        self.W = None

    def fit(self, X, y):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
                   Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return: self
        """
        return self.partial_fit(X, y)

    def predict(self, X):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
           Input samples.
        :return: numpy.ndarray of shape (n_samples)
            Predictions for the samples in X.
        """
        ret = []
        X = np.c_[X, np.ones(X.shape[0])]
        for x in range(0, X.shape[0]):
            if self.W[-1] @ X[x] >= 0:
                ret.append(1)
            else:
                ret.append(-1)
        return np.array(ret)

    def get_params(self, deep=False):
        return {"T": self.T, "lam": self.lam}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self

    def partial_fit(self, X, y, **kwargs):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
                   Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return: self
        """
        y[y == 0] = -1
        X = np.c_[X, np.ones(X.shape[0])]
        self.W = np.zeros((self.T + 1, X.shape[1]))
        for t in range(0, self.T):
            ith = np.random.randint(0, X.shape[0])
            lr = 1.0 / (self.lam * (t + 1))
            w = self.W[t]
            dot = w @ X[ith]
            test = y[ith]*dot
            if y[ith] * dot < 1:
                self.W[t + 1] = (1 - (lr * self.lam)) * w + lr * y[ith] * X[ith]
            else:
                self.W[t + 1] = (1 - (lr * self.lam)) * w
        return self



import copy
import importlib
import inspect
import sys
from typing import Literal
import numpy as np

import river
from river import metrics

from sail.common.progress_bar import SAILProgressBar


class SAILModelScorer:
    def __init__(
        self,
        scoring: str | None = None,
        estimator_type: Literal["regressor", "classifier", "clusterer"] | None = None,
        sample_weight: float = 1.0,
        pipeline_mode: bool = False,
    ) -> None:
        self.scoring = scoring
        self.estimator_type = estimator_type
        self.sample_weight = sample_weight
        self.pipeline_mode = pipeline_mode
        self._metric = self._resolve_scoring(scoring, estimator_type)
        self._scorer_type = "Pipeline" if self.pipeline_mode else "Model"

    @property
    def get_progressive_score(self):
        return self._metric.get()

    def get_state(self):
        if hasattr(self, "_y_true"):
            return {"y_true": self._y_true, "y_pred": self._y_pred}
        else:
            return None

    def set_state(self, state):
        self.progressive_score(state["y_true"], state["y_pred"], verbose=1)

    def get_default_scorer(self, estimator_type):
        if estimator_type == "classifier":
            return metrics.Accuracy()
        elif estimator_type == "regressor":
            return metrics.R2()
        elif estimator_type == "clusterer":
            return metrics.Completeness()
        else:
            raise Exception(
                f"Invalid Estimator type. Estimator can only be a regressor, classifier or clusterer. Given estimator type: {estimator_type}."
            )

    def _resolve_scoring(self, scoring, estimator_type):
        assert not (
            scoring == estimator_type == None
        ), "Either scoring or estimator_type must be a non null value."

        if scoring is None:
            assert (
                estimator_type != "passthrough",
            ), "Scoring cannot be None when the estimator_type is set to passthrough"
            return self.get_default_scorer(estimator_type)

        try:
            if isinstance(scoring, str):
                module = importlib.import_module("river.metrics")
                _scoring_class = getattr(module, scoring)
                return _scoring_class()
            elif isinstance(scoring, metrics.base.Metric):
                return scoring
            elif inspect.isclass(scoring):
                _scoring_class = scoring
                valid_classes = [
                    class_name
                    for _, class_name in inspect.getmembers(
                        sys.modules["river.metrics"], inspect.isclass
                    )
                ]
                if _scoring_class in valid_classes:
                    return _scoring_class()
                else:
                    raise Exception
            else:
                raise Exception

        except:
            method_name = (
                scoring.__name__
                if inspect.isclass(scoring)
                else scoring
                if isinstance(scoring, str)
                else scoring.__class__.__name__
            )
            raise AttributeError(
                f"Method '{method_name}' is not available in river.metrics. Scoring must be a str or an instance of the {river.metrics.__all__}."
            )

    def score(self, y_true, y_pred, sample_weight=None, verbose=1):
        with SAILProgressBar(
            steps=len(y_pred),
            desc=f"SAIL {self._scorer_type} Score",
            params={
                "Metric": self._metric.__class__.__qualname__,
                "Batch Size": len(y_pred),
            },
            format="scoring",
            verbose=verbose,
        ) as progress:
            scorer = self._resolve_scoring(self.scoring, self.estimator_type)
            if not sample_weight:
                sample_weight = self.sample_weight
            for v1, v2 in zip(y_true, y_pred):
                scorer.update(v1, v2, sample_weight)
                progress.update()
            progress.update_params("Score", scorer.get())
        return scorer.get()

    def progressive_score(
        self, y_true, y_pred, sample_weight=None, detached=False, verbose=1
    ):
        with SAILProgressBar(
            steps=len(y_pred),
            desc=f"SAIL {self._scorer_type} Progressive Score",
            params={
                "Metric": self._metric.__class__.__qualname__,
                "Batch Size": len(y_pred),
            },
            format="scoring",
            verbose=verbose,
        ) as progress:
            if not hasattr(self, "_y_true"):
                self._y_true = y_true
                self._y_pred = y_pred
            else:
                self._y_true = np.hstack((self._y_true, y_true))
                self._y_pred = np.hstack((self._y_pred, y_pred))

            if not sample_weight:
                sample_weight = self.sample_weight

            if detached:
                scorer = copy.deepcopy(self._metric)
            else:
                scorer = self._metric

            for v1, v2 in zip(y_true, y_pred):
                scorer.update(v1, v2, sample_weight)
                progress.update()
            progress.update_params("P_Score", scorer.get())
        return scorer.get()

    def clear(self):
        self._metric = self._resolve_scoring(self.scoring, self.estimator_type)



"""
Base class for ensemble models
Based on: https://github.com/yzhao062/combo/blob/master/combo/models/base.py
and extended for distributed incremental models
"""

import warnings
from collections import defaultdict
from abc import ABC, abstractmethod
from sklearn.utils import column_or_1d
from sklearn.utils.multiclass import check_classification_targets
import numpy as np
from inspect import signature


class BaseAggregator(ABC):

    @abstractmethod
    def __init__(self, base_estimators, fitted_estimators=[], window_size=200, pre_fitted=False):
        """
        :param base_estimators: Estimator objects with partial_fit defined
                    (incremental learning models).
        :param fitted_estimators: Batch learning estimators.
                Used only for scoring.
        :param metrics: Object that defines the quality of predictions
           (ex. metrics.accuracy_score in scikit-learn)
        """
        assert (isinstance(base_estimators, (list)))
        assert (isinstance(fitted_estimators, (list)))
        if (len(base_estimators) + len(fitted_estimators)) < 2:
            raise ValueError('At least 2 estimators are required')
        self.base_estimators = base_estimators
        self.fitted_estimators = fitted_estimators
        self.n_base_estimators_ = len(self.base_estimators)
        self.n_fitted_estimators_ = len(self.fitted_estimators)
        self.window_size = window_size

    def fit(self, X, y=None):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return:
        """
        return self.partial_fit(X,y)

    @abstractmethod
    def partial_fit(self, X, y=None, classes=None):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param classes: numpy.ndarray, optional.
               Unique classes in the data y.
        :return:
        """
        pass

    @abstractmethod
    def predict(self, X):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :return: numpy array of shape (n_samples,).
             Class labels/predictions for input samples.
        """
        pass

    # @abstractmethod
    # def predict_proba(self, X):
    #     """Return probability estimates for the test data X.
    #     Parameters
    #     ----------
    #     X : numpy array of shape (n_samples, n_features)
    #         The input samples.
    #     Returns
    #     -------
    #     p : numpy array of shape (n_samples,)
    #         The class probabilities of the input samples.
    #         Classes are ordered by lexicographic order.
    #     """
    #     pass

    def _set_n_classes(self, y):
        self._classes = 2  # default as binary classification
        if y is not None:
            check_classification_targets(y)
            self._classes = len(np.unique(y))

        return self

    def _set_weights(self, weights):
        """Internal function to set estimator weights.
        Parameters
        ----------
        weights : numpy array of shape (n_estimators,)
            Estimator weights. May be used after the alignment.
        Returns
        -------
        self
        """

        if weights is None:
            self.weights = np.ones([1, self.n_base_estimators_])
        else:
            self.weights = column_or_1d(weights).reshape(1, len(weights))
            assert (self.weights.shape[1] == self.n_base_estimators_)

            # adjust probability by a factor for integrity （added to 1）
            adjust_factor = self.weights.shape[1] / np.sum(weights)
            self.weights = self.weights * adjust_factor

            print(self.weights)
        return self

    def __len__(self):
        """Returns the number of estimators in the ensemble."""
        return len(self.base_estimators)

    def __getitem__(self, index):
        """Returns the index'th estimator in the ensemble."""
        return self.base_estimators[index]

    def __iter__(self):
        """Returns iterator over estimators in the ensemble."""
        return iter(self.estimators)

    # noinspection PyMethodParameters
    def _get_param_names(cls):
        # noinspection PyPep8
        """Get parameter names for the estimator
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        """
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError("scikit-learn estimators should always "
                                   "specify their parameters in the signature"
                                   " of their __init__ (no varargs)."
                                   " %s with constructor %s doesn't "
                                   " follow this convention."
                                   % (cls, init_signature))
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    # noinspection PyPep8
    def get_params(self, deep=True):
        """Get parameters for this estimator.
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """

        out = dict()
        for key in self._get_param_names():
            # We need deprecation warnings to always be on in order to
            # catch deprecated param values.
            # This is set in utils/__init__.py but it gets overwritten
            # when running under python3 somehow.
            warnings.simplefilter("always", DeprecationWarning)
            try:
                with warnings.catch_warnings(record=True) as w:
                    value = getattr(self, key, None)
                if len(w) and w[0].category == DeprecationWarning:
                    # if the parameter is deprecated, don't show it
                    continue
            finally:
                warnings.filters.pop(0)

            # XXX: should we rather test if instance of estimator?
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        # noinspection PyPep8
        """Set the parameters of this estimator.
        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.
        See http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html
        and sklearn/base.py for more information.
        Returns
        -------
        self : object
        """

        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

    # @property
    def estimators(self):
        return self.base_estimators, self.fitted_estimators

    # @property
    def base_estimators(self):
        return self.base_estimators

    # @property
    def fitted_estimators(self):
        return self.fitted_estimators



import river.linear_model as linear_model
import typing
from river import optim
from sail.models.river.base import SailRiverClassifier, SailRiverRegressor

__all__ = [
    "ALMAClassifier",
    "PAClassifier",
    "PARegressor",
    "SoftmaxRegression",
    "LinearRegression",
    "LogisticRegression",
    "Perceptron",
]


class LogisticRegression(SailRiverClassifier):
    def __init__(
        self,
        optimizer: optim.base.Optimizer = None,
        loss: optim.losses.BinaryLoss = None,
        l2=0.0,
        l1=0.0,
        intercept_init=0.0,
        intercept_lr: typing.Union[float, optim.base.Scheduler] = 0.01,
        clip_gradient=1e12,
        initializer: optim.base.Initializer = None,
    ):
        super(LogisticRegression, self).__init__(
            river_estimator=linear_model.LogisticRegression(
                optimizer,
                loss,
                l2,
                l1,
                intercept_init,
                intercept_lr,
                clip_gradient,
                initializer,
            )
        )


class LinearRegression(SailRiverRegressor):
    def __init__(
        self,
        optimizer: optim.base.Optimizer = None,
        loss: optim.losses.RegressionLoss = None,
        l2=0.0,
        l1=0.0,
        intercept_init=0.0,
        intercept_lr: typing.Union[optim.base.Scheduler, float] = 0.01,
        clip_gradient=1e12,
        initializer: optim.base.Initializer = None,
    ):
        super(LinearRegression, self).__init__(
            river_estimator=linear_model.LinearRegression(
                optimizer,
                loss,
                l2,
                l1,
                intercept_init,
                intercept_lr,
                clip_gradient,
                initializer,
            )
        )


class Perceptron(SailRiverClassifier):
    def __init__(
        self,
        l2=0.0,
        clip_gradient=1e12,
        initializer: optim.initializers.Initializer = None,
    ):
        super(Perceptron, self).__init__(
            river_estimator=linear_model.Perceptron(
                l2,
                clip_gradient,
                initializer,
            )
        )


class ALMAClassifier(SailRiverClassifier):
    def __init__(self, p=2, alpha=0.9, B=1 / 0.9, C=2**0.5):
        super(ALMAClassifier, self).__init__(
            river_estimator=linear_model.alma.ALMAClassifier(
                p,
                alpha,
                B,
                C,
            )
        )


class PAClassifier(SailRiverClassifier):
    def __init__(self, C=1.0, mode=1, learn_intercept=True):
        super(PAClassifier, self).__init__(
            river_estimator=linear_model.pa.PAClassifier(C, mode, learn_intercept)
        )


class PARegressor(SailRiverRegressor):
    def __init__(self, C=1.0, mode=1, eps=0.1, learn_intercept=True):
        super(PARegressor, self).__init__(
            river_estimator=linear_model.pa.PARegressor(C, mode, eps, learn_intercept)
        )


class SoftmaxRegression(SailRiverClassifier):
    def __init__(
        self,
        optimizer: optim.base.Optimizer = None,
        loss: optim.losses.MultiClassLoss = None,
        l2=0,
    ):
        super(SoftmaxRegression, self).__init__(
            river_estimator=linear_model.softmax.SoftmaxRegression(optimizer, loss, l2)
        )



from typing import Union, List
import matplotlib
import matplotlib.pyplot as plt
import numpy


def plot_series(
    y_orig: Union[List[float], numpy.ndarray],
    y_pred: Union[List[float], numpy.ndarray],
    plot_title="Time Series Plot",
    xaxis_label="Timestep",
    yaxis_label="Value",
    ybound=None,
    save_path=None,
    ts_timestamp=[],
):
    """Plot time series. Plot lines marking each prediction time.

    Parameters
    ----------
    y_orig : Union[List[float], numpy.ndarray]
        Original time series values
    y_pred : Union[List[float], numpy.ndarray]
        Predicted time series values.
    plot_title : str, default=""
        Plot title
    xaxis_label : str, default="Timestep"
        _description_
    yaxis_label : str, default="Value"
        _description_
    ybound : list, default=None
        y-axis range for time series plot.
    save_path : str, default=None
        Location with file name to save the time series plot.
    ts_timestamp : list, default=[]
        Timestamp values to display on the axis.
    """

    total_predictions = len(y_pred)
    font_size = 20
    fig = plt.figure()
    matplotlib.rcParams.update({"font.size": font_size})
    figure_size = (15.4, 7)
    fig.set_size_inches(figure_size)

    plt.title(plot_title)
    plt.ylabel(yaxis_label, fontsize=font_size)
    if len(ts_timestamp) > 0:
        plt.xlabel("Timestamp", fontsize=font_size)
    else:
        plt.xlabel(xaxis_label, fontsize=font_size)

    plt.plot(
        list(range(0, total_predictions)),
        y_pred,
        label="Prediction",
        color="blue",
        linewidth=1.8,
    )
    plt.scatter(
        list(range(0, total_predictions)),
        y_orig,
        label="Truth",
        marker=".",  # type: ignore
        color="green",
        s=60,
    )

    if len(ts_timestamp) > 0:
        timestamp = [i for i in ts_timestamp]
        for i in timestamp:
            plt.axvline(x=i, color="k", ls="dashed", linewidth=2.0)

    plt.legend(loc="upper left", fontsize=25)
    axes = plt.gca()
    axes.set_xlim([1, total_predictions + 10])

    if ybound is None:
        ybound = [0, max(y_orig + y_pred)]
    assert (
        len(ybound) == 2
    ), "ybound must have a lower bound and an upper bound. Ex: [0, 6]"
    axes.set_ylim([ybound[0], ybound[1]])

    if save_path is not None:
        fig.savefig(
            save_path,
            bbox_inches="tight",
        )

    plt.show()



import os
import tensorflow.keras as keras
from sail.utils.serialization import save_obj, load_obj
from sail.utils.logging import configure_logger
from scikeras.wrappers import KerasRegressor

LOGGER = configure_logger(logger_name="TF_Keras")


class KerasSerializationMixin:
    def save(self, model_folder):
        """
        Saves Model to Tensorflow SavedModel format.

        Args:
            model_folder: String, PathLike, path to SavedModel.
        """
        if self.model_ is not None:
            metadata_path = os.path.join(model_folder, "metadata")
            self.model_.save(metadata_path)
            # self.save_metrics(model_folder)
            LOGGER.info("Model saved successfully.")

    def load(self, model_folder):
        """
        Load Model from the mentioned folder location

        Args:
            model_folder: One of the following:
                - String or `pathlib.Path` object, path to the saved model
        """
        metadata_path = os.path.join(model_folder, "metadata")
        if os.path.exists(metadata_path):
            self.model = keras.models.load_model(metadata_path)
            # self.model_.metrics = self.load_metrics(model_folder)
            LOGGER.info("Model loaded successfully.")

    # def save_metrics(self, model_folder: str):
    #     try:
    #         metrics_path = os.path.join(model_folder, "metrics")
    #         for index, metric in enumerate(self.model_.metrics):
    #             name = metric.name + "_" + str(index)
    #             save_obj(
    #                 obj={"class": metric.__class__, "config": metric.get_config()},
    #                 location=metrics_path,
    #                 file_name=name,
    #             )
    #             LOGGER.info("Metric %s saved successfully.." % (name))
    #     except Exception as e:
    #         LOGGER.info("Not able to save metrics")

    # def load_metrics(self, model_folder: str):
    #     metrics = []
    #     try:
    #         metrics_path = os.path.join(model_folder, "metrics")
    #         for filename in os.listdir(metrics_path):
    #             filename = filename.split(".")[0]
    #             metric_obj = load_obj(location=metrics_path, file_name=filename)
    #             metric = metric_obj["class"].from_config(metric_obj["config"])
    #             metrics.append(metric)
    #             LOGGER.info("Metric %s loaded successfully.." % (metric.name))
    #     except Exception as e:
    #         LOGGER.info("Not able to load metrics")
    #     return metrics



import river.preprocessing.feature_hasher as feature_hasher
import river.preprocessing.impute as impute
import river.preprocessing.lda as lda
import river.preprocessing.one_hot as one_hot
import river.preprocessing.scale as scale
from sail.transformers.river.base import BaseRiverTransformer
from sklearn.base import OneToOneFeatureMixin

__all__ = [
    "AdaptiveStandardScaler",
    "Binarizer",
    "FeatureHasher",
    "LDA",
    "MaxAbsScaler",
    "MinMaxScaler",
    "Normalizer",
    "OneHotEncoder",
    "PreviousImputer",
    "RobustScaler",
    "StandardScaler",
    "StatImputer",
]


class FeatureHasher(BaseRiverTransformer):
    def __init__(self, n_features=1048576, seed: int = None):
        super(FeatureHasher, self).__init__(
            river_estimator=feature_hasher.FeatureHasher(n_features, seed)
        )


class PreviousImputer(BaseRiverTransformer):
    def __init__(self):
        super(PreviousImputer, self).__init__(river_estimator=impute.PreviousImputer())


class StatImputer(BaseRiverTransformer):
    def __init__(self, *imputers):
        super(StatImputer, self).__init__(river_estimator=impute.StatImputer(*imputers))


class LDA(BaseRiverTransformer):
    def __init__(
        self,
        n_components=10,
        number_of_documents=1e6,
        alpha_theta=0.5,
        alpha_beta=100.0,
        tau=64.0,
        kappa=0.75,
        vocab_prune_interval=10,
        number_of_samples=10,
        ranking_smooth_factor=1e-12,
        burn_in_sweeps=5,
        maximum_size_vocabulary=4000,
        seed: int = None,
    ):
        super(LDA, self).__init__(
            river_estimator=lda.LDA(
                n_components,
                number_of_documents,
                alpha_theta,
                alpha_beta,
                tau,
                kappa,
                vocab_prune_interval,
                number_of_samples,
                ranking_smooth_factor,
                burn_in_sweeps,
                maximum_size_vocabulary,
                seed,
            )
        )


class OneHotEncoder(BaseRiverTransformer):
    def __init__(self, sparse=False):
        super(OneHotEncoder, self).__init__(
            river_estimator=one_hot.OneHotEncoder(sparse)
        )


class AdaptiveStandardScaler(OneToOneFeatureMixin, BaseRiverTransformer):
    def __init__(self, alpha=0.3):
        super(AdaptiveStandardScaler, self).__init__(
            river_estimator=scale.AdaptiveStandardScaler(alpha)
        )


class Binarizer(BaseRiverTransformer):
    def __init__(self, threshold=0.0, dtype=bool):
        super(Binarizer, self).__init__(
            river_estimator=scale.Binarizer(threshold, dtype)
        )


class MaxAbsScaler(BaseRiverTransformer):
    def __init__(self):
        super(MaxAbsScaler, self).__init__(river_estimator=scale.MaxAbsScaler())


class MinMaxScaler(BaseRiverTransformer):
    def __init__(self):
        super(MinMaxScaler, self).__init__(river_estimator=scale.MinMaxScaler())


class Normalizer(BaseRiverTransformer):
    def __init__(self, order=2):
        super(Normalizer, self).__init__(river_estimator=scale.Normalizer(order))


class RobustScaler(BaseRiverTransformer):
    def __init__(self, with_centering=True, with_scaling=True, q_inf=0.25, q_sup=0.75):
        super(RobustScaler, self).__init__(
            river_estimator=scale.RobustScaler(
                with_centering, with_scaling, q_inf, q_sup
            )
        )


class StandardScaler(OneToOneFeatureMixin, BaseRiverTransformer):
    def __init__(self, with_std=True):
        super(StandardScaler, self).__init__(
            river_estimator=scale.StandardScaler(with_std)
        )



import torch
import torch.nn as nn
from skorch.regressor import NeuralNetRegressor
from sail.models.torch.base import TorchSerializationMixin


class lightLSTM(nn.Module):
    def __init__(self, ni, nh, nlayers, no):
        super().__init__()
        self.hidden_size = nh
        self.lstm = nn.LSTM(ni, nh, nlayers, batch_first=True)
        self.fc = nn.Linear(nh, no)

    def forward(self, x, hs):
        out, hs = self.lstm(x, hs)  # out.shape = (batch_size, seq_len, hidden_size)
        out = out.view(-1, self.hidden_size)  # out.shape = (seq_len, hidden_size)
        out = self.fc(out)

        return out, hs


class LSTMRegressor(NeuralNetRegressor, TorchSerializationMixin):
    def __init__(self, ni, no, nh, nlayers, module=lightLSTM, **kwargs):
        super(LSTMRegressor, self).__init__(
            module=module,
            module__ni=ni,
            module__no=no,
            module__nh=nh,
            module__nlayers=nlayers,
            **kwargs
        )



"""
Omni-Scale 1D-CNN

Code adapted from https://github.com/Wensi-Tang/OS-CNN
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import math
import numpy as np
import torch.nn as nn 
import torch.optim as optim
from skorch.classifier import NeuralNetClassifier


def get_Prime_number_in_a_range(start, end):
    Prime_list = []
    for val in range(start, end + 1): 
        prime_or_not = True
        for n in range(2, val):
            if (val % n) == 0:
                prime_or_not = False
                break
        if prime_or_not:
            Prime_list.append(val)
    return Prime_list


def get_out_channel_number(paramenter_layer, in_channel, prime_list):
    out_channel_expect = int(paramenter_layer/(in_channel*sum(prime_list)))
    return out_channel_expect


def generate_layer_parameter_list(start,end,paramenter_number_of_layer_list, in_channel = 1):
    prime_list = get_Prime_number_in_a_range(start, end)
    # if prime_list == []:
    #     # print('start = ',start, 'which is larger than end = ', end)
    input_in_channel = in_channel
    layer_parameter_list = []
    for paramenter_number_of_layer in paramenter_number_of_layer_list:
        out_channel = get_out_channel_number(paramenter_number_of_layer, in_channel, prime_list)
        
        tuples_in_layer= []
        for prime in prime_list:
            tuples_in_layer.append((in_channel,out_channel,prime))
        in_channel =  len(prime_list)*out_channel
        
        layer_parameter_list.append(tuples_in_layer)
    
    tuples_in_layer_last = []
    first_out_channel = len(prime_list)*get_out_channel_number(paramenter_number_of_layer_list[0], input_in_channel, prime_list)
    tuples_in_layer_last.append((in_channel,first_out_channel,start))
    tuples_in_layer_last.append((in_channel,first_out_channel,start+1))
    layer_parameter_list.append(tuples_in_layer_last)
    return layer_parameter_list


def calculate_mask_index(kernel_length_now,largest_kernel_lenght):
    right_zero_mast_length = math.ceil((largest_kernel_lenght-1)/2)-math.ceil((kernel_length_now-1)/2)
    left_zero_mask_length = largest_kernel_lenght - kernel_length_now - right_zero_mast_length
    return left_zero_mask_length, left_zero_mask_length+ kernel_length_now


def creat_mask(number_of_input_channel,number_of_output_channel, kernel_length_now, largest_kernel_lenght):
    ind_left, ind_right= calculate_mask_index(kernel_length_now,largest_kernel_lenght)
    mask = np.ones((number_of_input_channel,number_of_output_channel,largest_kernel_lenght))
    mask[:,:,0:ind_left]=0
    mask[:,:,ind_right:]=0
    return mask


def creak_layer_mask(layer_parameter_list):
    largest_kernel_lenght = layer_parameter_list[-1][-1]
    mask_list = []
    init_weight_list = []
    bias_list = []
    for i in layer_parameter_list:
        conv = torch.nn.Conv1d(in_channels=i[0], out_channels=i[1], kernel_size=i[2])
        ind_l,ind_r= calculate_mask_index(i[2],largest_kernel_lenght)
        big_weight = np.zeros((i[1],i[0],largest_kernel_lenght))
        big_weight[:,:,ind_l:ind_r]= conv.weight.detach().numpy()
        
        bias_list.append(conv.bias.detach().numpy())
        init_weight_list.append(big_weight)
        
        mask = creat_mask(i[1],i[0],i[2], largest_kernel_lenght)
        mask_list.append(mask)
        
    mask = np.concatenate(mask_list, axis=0)
    init_weight = np.concatenate(init_weight_list, axis=0)
    init_bias = np.concatenate(bias_list, axis=0)
    return mask.astype(np.float32), init_weight.astype(np.float32), init_bias.astype(np.float32)

    
class build_layer_with_layer_parameter(nn.Module):
    def __init__(self,layer_parameters):
        super(build_layer_with_layer_parameter, self).__init__()

        os_mask, init_weight, init_bias= creak_layer_mask(layer_parameters)
        
        
        in_channels = os_mask.shape[1] 
        out_channels = os_mask.shape[0] 
        max_kernel_size = os_mask.shape[-1]

        self.weight_mask = nn.Parameter(torch.from_numpy(os_mask),requires_grad=False)
        
        self.padding = nn.ConstantPad1d((int((max_kernel_size-1)/2), int(max_kernel_size/2)), 0)
         
        self.conv1d = torch.nn.Conv1d(in_channels=in_channels, out_channels=out_channels, kernel_size=max_kernel_size)
        self.conv1d.weight = nn.Parameter(torch.from_numpy(init_weight),requires_grad=True)
        self.conv1d.bias =  nn.Parameter(torch.from_numpy(init_bias),requires_grad=True)

        self.bn = nn.BatchNorm1d(num_features=out_channels)
    
    def forward(self, X):
        self.conv1d.weight.data = self.conv1d.weight*self.weight_mask
        #self.conv1d.weight.data.mul_(self.weight_mask)
        result_1 = self.padding(X)
        result_2 = self.conv1d(result_1)
        result_3 = self.bn(result_2)
        result = F.relu(result_3)
        return result    
    
    
class _OS_CNN(nn.Module):
    def __init__(
        self,
        n_class, #number of classes to predict
        input_channel,
        receptive_field_shape,
        start_kernel_size = 1,
        max_kernel_size = 89,
        paramenter_number_of_layer_list = [8*128, 5*128*256 + 2*256*128],
        few_shot = False
    ):
        super(_OS_CNN, self).__init__()
        self.few_shot = few_shot
        self.start_kernel_size = start_kernel_size
        self.max_kernel_size = max_kernel_size
        self.paramenter_number_of_layer_list = paramenter_number_of_layer_list
        self.input_channel = input_channel
        self.receptive_field_shape = receptive_field_shape
        
        layer_parameter_list = generate_layer_parameter_list(self.start_kernel_size,
                                                     receptive_field_shape,
                                                     self.paramenter_number_of_layer_list,
                                                     in_channel = self.input_channel)
        self.layer_list = []
        
        # building each model layer from a list of parameter
        for i in range(len(layer_parameter_list)):
            layer = build_layer_with_layer_parameter(layer_parameter_list[i])
            
            # print([layer.shape for layer in layer.parameters()])
            self.layer_list.append(layer)
        
        self.net = nn.Sequential(*self.layer_list)
            
        self.averagepool = nn.AdaptiveAvgPool1d(1)
        
        out_put_channel_numebr = 0
        for final_layer_parameters in layer_parameter_list[-1]:
            out_put_channel_numebr = out_put_channel_numebr+ final_layer_parameters[1] 
            
        self.hidden = nn.Linear(out_put_channel_numebr, n_class)

    def forward(self, X):
        
        X = self.net(X)

        X = self.averagepool(X)
        X = X.squeeze_(-1)

        if not self.few_shot:
            X = self.hidden(X)
        return X
    

class OS_CNN_CLassifier(NeuralNetClassifier):
    """
    Basic OS_CNN model
    
    Args:
        n_class: Number of classes to predict]
        input_channel: Number of input channels. varies depending on data type (univariate and multivariate)
        receptive_field_shape: Scales of receptive fields to avoid scale tuning 
        start_kernel_size: Start kernel size. Must be a prime number 
        max_kernel_size: Maximum kernel size. Must be a prime number
        paramenter_number_of_layer_list: OS block layer size
        device: cpu or gpu
        max_epochs: Epoch number
        batch_size: Size of the batch
        criterion: Loss function
        optimizer: Optimizer
    
    """
    def __init__(
        self,
        n_class, 
        input_channel,
        receptive_field_shape,
        start_kernel_size = 1,
        max_kernel_size = 89,
        paramenter_number_of_layer_list = [8*128, 5*128*256 + 2*256*128],
        device = "cpu",
        few_shot = False,
        max_epochs = 20,
        **kwargs):
        super(OS_CNN_CLassifier, self).__init__(
            module = _OS_CNN,
            module__n_class = n_class,
            module__input_channel = input_channel,
            module__receptive_field_shape = receptive_field_shape,
            module__start_kernel_size = start_kernel_size,
            module__max_kernel_size = max_kernel_size,
            module__paramenter_number_of_layer_list = paramenter_number_of_layer_list,
            module__few_shot = few_shot,
            device = device,
            max_epochs=max_epochs, 
            batch_size=16,
            criterion = nn.CrossEntropyLoss,
            optimizer = optim.Adam,
            **kwargs
        )
        


from sail.utils.ray_utils import _model_fit
import ray
from sklearn.metrics import r2_score
from abc import ABC, abstractmethod
from sail.models.native.base import MetaEstimatorMixin

# Reference: https://github.com/AlexImb/automl-streams
# https://github.com/yzhao062/combo/blob/master/combo/models/base.py


class ModelSelectorBase(ABC, MetaEstimatorMixin):
    """
    Base class for distributed model selection.
    """

    def __init__(self, estimators, fitted_estimators=[], metrics=r2_score):
        """
        :param estimators: Estimator objects with partial_fit defined
                    (incremental learning models).
        :param fitted_estimators: Batch learning estimators.
                Used only for scoring.
        :param metrics: Object that defines the quality of predictions
           (ex. metrics.accuracy_score in scikit-learn)
        """
        self.base_estimators = estimators  # [e.__class__() for e in estimators]
        self.fitted_estimators = fitted_estimators
        self.best_model_index = 0
        self.metrics = metrics

    @abstractmethod
    def partial_fit(self, X, y=None, classes=None):
        """
        :param model: Any machine learning model with partial_fit function defined.
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param classes: numpy.ndarray, optional.
               Unique classes in the data y.
        :return:
        """
        raise NotImplementedError

    def _partial_fit_estimators(self, X, y, classes=None, sample_weight=None):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param classes: numpy.ndarray, optional.
               Unique classes in the data y.
        :param sample_weight: User defined weights for base estimators [Not implemented yet]
        :return: None
        """
        trained_models = []
        for estimator in self.base_estimators:
            estimator.n_features_in_ = X.shape[1]
            trained_models.append(
                _model_fit.remote(model=estimator, X=X, y=y, classes=classes)
            )
        self.base_estimators = ray.get(trained_models)

    @abstractmethod
    def _get_best_model_index(self, X, y):
        pass

    def fit(self, X, y, classes=None):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param classes: numpy.ndarray, optional.
               Unique classes in the data y.
        :return:
        """
        self.partial_fit(X, y, classes)

    def get_best_model_index(self, X, y):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels (ground truth) for the target variable.
        :return: int. Returns the index of the best model based on user defined metrics.
        """
        self.best_model_index = self._get_best_model_index(X, y)
        return self.best_model_index

    def predict_proba(self, X):
        """
        Return probability estimates for the test data X.
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :return: numpy array of shape (n_samples,)
            The class probabilities of the input samples.
        """
        estimators = self.base_estimators
        estimators.extend(self.fitted_estimators)
        return estimators[self.best_model_index].predict_proba(X)

    def predict(self, X):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :return: numpy array of shape (n_samples,).
             Class labels/predictions for input samples.
        """
        estimators = self.base_estimators
        estimators.extend(self.fitted_estimators)
        return estimators[self.best_model_index].predict(X)

    # def reset(self):
    #     self.base_estimators = [est.reset() for est in self.base_estimators]
    #     self.fitted_estimators = [est.reset() for est in self.fitted_estimators]
    #     self.best_model_index = 0
    #     return self

    def get_best_model(self):
        """
        :return: Current best model in the list of base estimators and fitted estimators.
        """
        estimators = self.base_estimators
        estimators.extend(self.fitted_estimators)
        return estimators[self.best_model_index]



from sklearn.linear_model import Perceptron
from sklearn.linear_model import SGDClassifier
from sklearn.linear_model import PassiveAggressiveClassifier
from sklearn.linear_model import SGDRegressor
from sklearn.linear_model import PassiveAggressiveRegressor

__all__ = [
    "Perceptron",
    "SGDClassifier",
    "PassiveAggressiveClassifier",
    "SGDRegressor",
    "PassiveAggressiveRegressor"
]


import typing

from river.forest import ARFClassifier, ARFRegressor
from river import base, metrics
from river.tree.splitter import Splitter
from sail.models.river.base import SailRiverClassifier, SailRiverRegressor

__all__ = [
    "AdaptiveRandomForestClassifier",
    "AdaptiveRandomForestRegressor",
]


class AdaptiveRandomForestClassifier(SailRiverClassifier):
    def __init__(
        self,
        n_models: int = 10,
        max_features: bool | str | int = "sqrt",
        lambda_value: int = 6,
        metric: metrics.base.MultiClassMetric | None = None,
        disable_weighted_vote=False,
        drift_detector: base.DriftDetector | None = None,
        warning_detector: base.DriftDetector | None = None,
        # Tree parameters
        grace_period: int = 50,
        max_depth: int | None = None,
        split_criterion: str = "info_gain",
        delta: float = 0.01,
        tau: float = 0.05,
        leaf_prediction: str = "nba",
        nb_threshold: int = 0,
        nominal_attributes: list | None = None,
        splitter: Splitter | None = None,
        binary_split: bool = False,
        min_branch_fraction: float = 0.01,
        max_share_to_split: float = 0.99,
        max_size: float = 100.0,
        memory_estimate_period: int = 2_000_000,
        stop_mem_management: bool = False,
        remove_poor_attrs: bool = False,
        merit_preprune: bool = True,
        seed: int | None = None,
    ):
        super(AdaptiveRandomForestClassifier, self).__init__(
            river_estimator=ARFClassifier(
                n_models,
                max_features,
                lambda_value,
                metric,
                disable_weighted_vote,
                drift_detector,
                warning_detector,
                grace_period,
                max_depth,
                split_criterion,
                delta,
                tau,
                leaf_prediction,
                nb_threshold,
                nominal_attributes,
                splitter,
                binary_split,
                min_branch_fraction,
                max_share_to_split,
                max_size,
                memory_estimate_period,
                stop_mem_management,
                remove_poor_attrs,
                merit_preprune,
                seed,
            )
        )


class AdaptiveRandomForestRegressor(SailRiverRegressor):
    def __init__(
        self,
        # Forest parameters
        n_models: int = 10,
        max_features="sqrt",
        aggregation_method: str = "median",
        lambda_value: int = 6,
        metric: metrics.base.RegressionMetric | None = None,
        disable_weighted_vote=True,
        drift_detector: base.DriftDetector | None = None,
        warning_detector: base.DriftDetector | None = None,
        # Tree parameters
        grace_period: int = 50,
        max_depth: int | None = None,
        delta: float = 0.01,
        tau: float = 0.05,
        leaf_prediction: str = "adaptive",
        leaf_model: base.Regressor | None = None,
        model_selector_decay: float = 0.95,
        nominal_attributes: list | None = None,
        splitter: Splitter | None = None,
        min_samples_split: int = 5,
        binary_split: bool = False,
        max_size: float = 500.0,
        memory_estimate_period: int = 2_000_000,
        stop_mem_management: bool = False,
        remove_poor_attrs: bool = False,
        merit_preprune: bool = True,
        seed: int | None = None,
    ):
        super(AdaptiveRandomForestRegressor, self).__init__(
            river_estimator=ARFRegressor(
                n_models,
                max_features,
                aggregation_method,
                lambda_value,
                metric,
                disable_weighted_vote,
                drift_detector,
                warning_detector,
                grace_period,
                max_depth,
                delta,
                tau,
                leaf_prediction,
                leaf_model,
                model_selector_decay,
                nominal_attributes,
                splitter,
                min_samples_split,
                binary_split,
                max_size,
                memory_estimate_period,
                stop_mem_management,
                remove_poor_attrs,
                merit_preprune,
                seed,
            )
        )
